# Copyright (C) 2007-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
# cython: language_level=3

"""Helper functions for DirState.

This is the python implementation for DirState functions.
"""

import binascii
import bisect
import codecs
import errno
import os
import stat
import sys

from .errors import BzrFormatsError, BadFileKindError
from . import osutils
from .osutils import (is_inside, is_inside_any, parent_directories, pathjoin,
                     splitpath, file_kind_from_stat_mode, sha_file, _walkdirs_utf8)
from .dirstate import DirstateInventoryChange

# Delay import to avoid circular dependency
DirstateCorrupt = None
DirState = None

def _ensure_dirstate_import():
    global DirstateCorrupt, DirState
    if DirstateCorrupt is None:
        from .dirstate import DirstateCorrupt as DC, DirState as DS
        DirstateCorrupt = DC
        DirState = DS

from cpython.tuple cimport PyTuple_New, PyTuple_SET_ITEM


# This is the Windows equivalent of ENOTDIR
# It is defined in pywin32.winerror, but we don't want a strong dependency for
# just an error code.
# XXX: Perhaps we could get it from a windows header ?
cdef int ERROR_PATH_NOT_FOUND
ERROR_PATH_NOT_FOUND = 3
cdef int ERROR_DIRECTORY
ERROR_DIRECTORY = 267

cdef extern from "python-compat.h":
    unsigned long htonl(unsigned long)

# Give Pyrex some function definitions for it to understand.
# All of these are just hints to Pyrex, so that it can try to convert python
# objects into similar C objects. (such as PyInt => int).
# In anything defined 'cdef extern from XXX' the real C header will be
# imported, and the real definition will be used from there. So these are just
# hints, and do not need to match exactly to the C definitions.

cdef extern from *:
    ctypedef unsigned long size_t

cdef extern from "stdint.h":
    ctypedef long intptr_t


cdef extern from "stdlib.h":
    unsigned long int strtoul(char *nptr, char **endptr, int base)


cdef extern from 'sys/stat.h':
    int S_ISDIR(int mode)
    int S_ISREG(int mode)
    # On win32, this actually comes from "python-compat.h"
    int S_ISLNK(int mode)
    int S_IXUSR

# These functions allow us access to a bit of the 'bare metal' of python
# objects, rather than going through the object abstraction. (For example,
# PyList_Append, rather than getting the 'append' attribute of the object, and
# creating a tuple, and then using PyCallObject).
# Functions that return (or take) a void* are meant to grab a C PyObject*. This
# differs from the Pyrex 'object'. If you declare a variable as 'object' Pyrex
# will automatically Py_INCREF and Py_DECREF when appropriate. But for some
# inner loops, we don't need to do that at all, as the reference only lasts for
# a very short time.
# Note that the C API GetItem calls borrow references, so pyrex does the wrong
# thing if you declare e.g. object PyList_GetItem(object lst, int index) - you
# need to manually Py_INCREF yourself.
cdef extern from "Python.h":
    ctypedef int Py_ssize_t
    ctypedef struct PyObject:
        pass
    int PyList_Append(object lst, object item) except -1
    void *PyList_GetItem_object_void "PyList_GET_ITEM" (object lst, int index)
    void *PyList_GetItem_void_void "PyList_GET_ITEM" (void * lst, int index)
    object PyList_GET_ITEM(object lst, Py_ssize_t index)
    int PyList_CheckExact(object)
    Py_ssize_t PyList_GET_SIZE (object p)

    void *PyTuple_GetItem_void_void "PyTuple_GET_ITEM" (void* tpl, int index)
    object PyTuple_GetItem_void_object "PyTuple_GET_ITEM" (void* tpl, int index)
    object PyTuple_GET_ITEM(object tpl, Py_ssize_t index)

    unsigned long PyLong_AsUnsignedLongMask(object number) except? -1

    char *PyBytes_AsString(object p)
    char *PyBytes_AsString_obj "PyBytes_AsString" (PyObject *string)
    char *PyBytes_AS_STRING_void "PyBytes_AS_STRING" (void *p)
    int PyBytes_AsStringAndSize(object str, char **buffer, Py_ssize_t *length) except -1
    object PyBytes_FromString(char *)
    object PyBytes_FromStringAndSize(char *, Py_ssize_t)
    int PyBytes_Size(object p)
    int PyBytes_GET_SIZE_void "PyBytes_GET_SIZE" (void *p)
    int PyBytes_CheckExact(object p)
    int PyFloat_Check(object p)
    double PyFloat_AsDouble(object p)
    int PyLong_Check(object p)
    void Py_INCREF(object o)
    void Py_DECREF(object o)


cdef extern from "string.h":
    int strncmp(char *s1, char *s2, int len)
    void *memchr(void *s, int c, size_t len)
    int memcmp(void *b1, void *b2, size_t len)

from ._str_helpers cimport _my_memrchr, safe_string_from_size


cdef int _is_aligned(void *ptr): # cannot_raise
    """Is this pointer aligned to an integer size offset?

    :return: 1 if this pointer is aligned, 0 otherwise.
    """
    return ((<intptr_t>ptr) & ((sizeof(int))-1)) == 0


cdef int _cmp_by_dirs(char *path1, int size1, char *path2, int size2): # cannot_raise
    cdef unsigned char *cur1
    cdef unsigned char *cur2
    cdef unsigned char *end1
    cdef unsigned char *end2
    cdef int *cur_int1
    cdef int *cur_int2
    cdef int *end_int1
    cdef int *end_int2

    if path1 == path2 and size1 == size2:
        return 0

    end1 = <unsigned char*>path1+size1
    end2 = <unsigned char*>path2+size2

    # Use 32-bit comparisons for the matching portion of the string.
    # Almost all CPU's are faster at loading and comparing 32-bit integers,
    # than they are at 8-bit integers.
    # 99% of the time, these will be aligned, but in case they aren't just skip
    # this loop
    if _is_aligned(path1) and _is_aligned(path2):
        cur_int1 = <int*>path1
        cur_int2 = <int*>path2
        end_int1 = <int*>(path1 + size1 - (size1 % sizeof(int)))
        end_int2 = <int*>(path2 + size2 - (size2 % sizeof(int)))

        while cur_int1 < end_int1 and cur_int2 < end_int2:
            if cur_int1[0] != cur_int2[0]:
                break
            cur_int1 = cur_int1 + 1
            cur_int2 = cur_int2 + 1

        cur1 = <unsigned char*>cur_int1
        cur2 = <unsigned char*>cur_int2
    else:
        cur1 = <unsigned char*>path1
        cur2 = <unsigned char*>path2

    while cur1 < end1 and cur2 < end2:
        if cur1[0] == cur2[0]:
            # This character matches, just go to the next one
            cur1 = cur1 + 1
            cur2 = cur2 + 1
            continue
        # The current characters do not match
        if cur1[0] == b'/':
            return -1 # Reached the end of path1 segment first
        elif cur2[0] == b'/':
            return 1 # Reached the end of path2 segment first
        elif cur1[0] < cur2[0]:
            return -1
        else:
            return 1

    # We reached the end of at least one of the strings
    if cur1 < end1:
        return 1 # Not at the end of cur1, must be at the end of cur2
    if cur2 < end2:
        return -1 # At the end of cur1, but not at cur2
    # We reached the end of both strings
    return 0


cdef class Reader:
    """Maintain the current location, and return fields as you parse them."""

    cdef object state # The DirState object
    cdef object text # The overall string object
    cdef char *text_cstr # Pointer to the beginning of text
    cdef int text_size # Length of text

    cdef char *end_cstr # End of text
    cdef char *cur_cstr # Pointer to the current record
    cdef char *next # Pointer to the end of this record

    def __init__(self, text, state):
        _ensure_dirstate_import()
        self.state = state
        self.text = text
        self.text_cstr = PyBytes_AsString(text)
        self.text_size = PyBytes_Size(text)
        self.end_cstr = self.text_cstr + self.text_size
        self.cur_cstr = self.text_cstr

    cdef char *get_next(self, int *size) except NULL:
        """Return a pointer to the start of the next field."""
        cdef char *next
        cdef Py_ssize_t extra_len

        if self.cur_cstr == NULL:
            raise AssertionError('get_next() called when cur_str is NULL')
        elif self.cur_cstr >= self.end_cstr:
            raise AssertionError('get_next() called when there are no chars'
                                 ' left')
        next = self.cur_cstr
        self.cur_cstr = <char*>memchr(next, b'\0', self.end_cstr - next)
        if self.cur_cstr == NULL:
            extra_len = self.end_cstr - next
            raise DirstateCorrupt(self.state,
                'failed to find trailing NULL (\\0).'
                ' Trailing garbage: %r'
                % safe_string_from_size(next, extra_len))
        size[0] = self.cur_cstr - next
        self.cur_cstr = self.cur_cstr + 1
        return next

    cdef object get_next_str(self):
        """Get the next field as a Python string."""
        cdef int size
        cdef char *next
        next = self.get_next(&size)
        return safe_string_from_size(next, size)

    cdef int _init(self) except -1:
        """Get the pointer ready.

        This assumes that the dirstate header has already been read, and we
        already have the dirblock string loaded into memory.
        This just initializes our memory pointers, etc for parsing of the
        dirblock string.
        """
        cdef char *first
        cdef int size
        # The first field should be an empty string left over from the Header
        first = self.get_next(&size)
        if first[0] != b'\0' and size == 0:
            raise AssertionError('First character should be null not: %s'
                                 % (first,))
        return 0

    cdef object _get_entry(self, int num_trees, void **p_current_dirname,
                           int *new_block):
        """Extract the next entry.

        This parses the next entry based on the current location in
        ``self.cur_cstr``.
        Each entry can be considered a "row" in the total table. And each row
        has a fixed number of columns. It is generally broken up into "key"
        columns, then "current" columns, and then "parent" columns.

        :param num_trees: How many parent trees need to be parsed
        :param p_current_dirname: A pointer to the current PyBytes
            representing the directory name.
            We pass this in as a void * so that pyrex doesn't have to
            increment/decrement the PyObject reference counter for each
            _get_entry call.
            We use a pointer so that _get_entry can update it with the new
            value.
        :param new_block: This is to let the caller know that it needs to
            create a new directory block to store the next entry.
        """
        cdef tuple path_name_file_id_key
        cdef tuple tmp
        cdef char *entry_size_cstr
        cdef unsigned long int entry_size
        cdef char* executable_cstr
        cdef int is_executable
        cdef char* dirname_cstr
        cdef char* trailing
        cdef int cur_size
        cdef int i
        cdef object minikind
        cdef object fingerprint
        cdef object info

        # Read the 'key' information (dirname, name, file_id)
        dirname_cstr = self.get_next(&cur_size)
        # Check to see if we have started a new directory block.
        # If so, then we need to create a new dirname PyBytes, so that it can
        # be used in all of the tuples. This saves time and memory, by re-using
        # the same object repeatedly.

        # Do the cheap 'length of string' check first. If the string is a
        # different length, then we *have* to be a different directory.
        if (cur_size != PyBytes_GET_SIZE_void(p_current_dirname[0])
            or strncmp(dirname_cstr,
                       # Extract the char* from our current dirname string.  We
                       # know it is a PyBytes, so we can use
                       # PyBytes_AS_STRING, we use the _void version because
                       # we are tricking Pyrex by using a void* rather than an
                       # <object>
                       PyBytes_AS_STRING_void(p_current_dirname[0]),
                       cur_size+1) != 0):
            dirname = safe_string_from_size(dirname_cstr, cur_size)
            p_current_dirname[0] = <void*>dirname
            new_block[0] = 1
        else:
            new_block[0] = 0

        # Build up the key that will be used.
        # By using <object>(void *) Pyrex will automatically handle the
        # Py_INCREF that we need.
        cur_dirname = <object>p_current_dirname[0]
        tmp = PyTuple_New(3)
        Py_INCREF(cur_dirname); PyTuple_SET_ITEM(tmp, 0, cur_dirname)
        cur_basename = self.get_next_str()
        cur_file_id = self.get_next_str()
        Py_INCREF(cur_basename); PyTuple_SET_ITEM(tmp, 1, cur_basename)
        Py_INCREF(cur_file_id); PyTuple_SET_ITEM(tmp, 2, cur_file_id)
        path_name_file_id_key = tmp

        # Parse all of the per-tree information. current has the information in
        # the same location as parent trees. The only difference is that 'info'
        # is a 'packed_stat' for current, while it is a 'revision_id' for
        # parent trees.
        # minikind, fingerprint, and info will be returned as regular python
        # strings
        # entry_size and is_executable will be parsed into a python Long and
        # python Boolean, respectively.
        # TODO: jam 20070718 Consider changin the entry_size conversion to
        #       prefer python Int when possible. They are generally faster to
        #       work with, and it will be rare that we have a file >2GB.
        #       Especially since this code is pretty much fixed at a max of
        #       4GB.
        trees = []
        for i from 0 <= i < num_trees:
            minikind = self.get_next_str()
            fingerprint = self.get_next_str()
            entry_size_cstr = self.get_next(&cur_size)
            entry_size = strtoul(entry_size_cstr, NULL, 10)
            executable_cstr = self.get_next(&cur_size)
            is_executable = (executable_cstr[0] == b'y')
            info = self.get_next_str()
            PyList_Append(trees, (
                minikind,     # minikind
                fingerprint,  # fingerprint
                entry_size,   # size
                is_executable,# executable
                info,         # packed_stat or revision_id
            ))

        # The returned tuple is (key, [trees])
        ret = (path_name_file_id_key, trees)
        # Ignore the trailing newline, but assert that it does exist, this
        # ensures that we always finish parsing a line on an end-of-entry
        # marker.
        trailing = self.get_next(&cur_size)
        if cur_size != 1 or not trailing.startswith(b'\n'):
            raise DirstateCorrupt(self.state,
                'Bad parse, we expected to end on \\n, not: %d %s: %s'
                % (cur_size, safe_string_from_size(trailing, cur_size),
                   ret))
        return ret

    def _parse_dirblocks(self):
        """Parse all dirblocks in the state file."""
        cdef int num_trees
        cdef object current_block
        cdef object entry
        cdef void * current_dirname
        cdef int new_block
        cdef int expected_entry_count
        cdef int entry_count

        num_trees = self.state._num_present_parents() + 1
        expected_entry_count = self.state._num_entries

        # Ignore the first record
        self._init()

        current_block = []
        dirblocks = [(b'', current_block), (b'', [])]
        self.state._dirblocks = dirblocks
        obj = b''
        current_dirname = <void*>obj
        new_block = 0
        entry_count = 0

        # TODO: jam 2007-05-07 Consider pre-allocating some space for the
        #       members, and then growing and shrinking from there. If most
        #       directories have close to 10 entries in them, it would save a
        #       few mallocs if we default our list size to something
        #       reasonable. Or we could malloc it to something large (100 or
        #       so), and then truncate. That would give us a malloc + realloc,
        #       rather than lots of reallocs.
        while self.cur_cstr < self.end_cstr:
            entry = self._get_entry(num_trees, &current_dirname, &new_block)
            if new_block:
                # new block - different dirname
                current_block = []
                PyList_Append(dirblocks,
                              (<object>current_dirname, current_block))
            PyList_Append(current_block, entry)
            entry_count = entry_count + 1
        if entry_count != expected_entry_count:
            raise DirstateCorrupt(self.state,
                    'We read the wrong number of entries.'
                    ' We expected to read %s, but read %s'
                    % (expected_entry_count, entry_count))
        self.state._split_root_dirblock_into_contents()


def _read_dirblocks(state):
    """Read in the dirblocks for the given DirState object.

    This is tightly bound to the DirState internal representation. It should be
    thought of as a member function, which is only separated out so that we can
    re-write it in pyrex.

    :param state: A DirState object.
    :return: None
    :postcondition: The dirblocks will be loaded into the appropriate fields in
        the DirState object.
    """
    state._state_file.seek(state._end_of_header)
    text = state._state_file.read()
    # TODO: check the crc checksums. crc_measured = zlib.crc32(text)

    reader = Reader(text, state)

    reader._parse_dirblocks()
    state._dirblock_state = DirState.IN_MEMORY_UNMODIFIED


cdef int minikind_from_mode(int mode): # cannot_raise
    # in order of frequency:
    if S_ISREG(mode):
        return c"f"
    if S_ISDIR(mode):
        return c"d"
    if S_ISLNK(mode):
        return c"l"
    return 0


_encode = binascii.b2a_base64


cdef unsigned long _time_to_unsigned(object t):  # cannot_raise
    if PyFloat_Check(t):
        t = t.__int__()
    return PyLong_AsUnsignedLongMask(t)


cdef _pack_stat(stat_value):
    """return a string representing the stat value's key fields.

    :param stat_value: A stat oject with st_size, st_mtime, st_ctime, st_dev,
        st_ino and st_mode fields.
    """
    cdef char result[6*4] # 6 long ints
    cdef int *aliased
    aliased = <int *>result
    aliased[0] = htonl(PyLong_AsUnsignedLongMask(stat_value.st_size))
    # mtime and ctime will often be floats but get converted to PyInt within
    aliased[1] = htonl(_time_to_unsigned(stat_value.st_mtime))
    aliased[2] = htonl(_time_to_unsigned(stat_value.st_ctime))
    aliased[3] = htonl(PyLong_AsUnsignedLongMask(stat_value.st_dev))
    aliased[4] = htonl(PyLong_AsUnsignedLongMask(stat_value.st_ino))
    aliased[5] = htonl(PyLong_AsUnsignedLongMask(stat_value.st_mode))
    packed = PyBytes_FromStringAndSize(result, 6*4)
    return _encode(packed)[:-1]


cpdef update_entry(self, entry, abspath, stat_value):
    """Update the entry based on what is actually on disk.

    This function only calculates the sha if it needs to - if the entry is
    uncachable, or clearly different to the first parent's entry, no sha
    is calculated, and None is returned.

    :param self: The dirstate object this is operating on.
    :param entry: This is the dirblock entry for the file in question.
    :param abspath: The path on disk for this file.
    :param stat_value: The stat value done on the path.
    :return: None, or The sha1 hexdigest of the file (40 bytes) or link
        target of a symlink.
    """
    # TODO - require pyrex 0.9.8, then use a pyd file to define access to the
    # _st mode of the compiled stat objects.
    cdef int minikind, saved_minikind
    cdef void * details
    cdef int worth_saving
    minikind = minikind_from_mode(stat_value.st_mode)
    if 0 == minikind:
        return None
    packed_stat = _pack_stat(stat_value)
    details = PyList_GetItem_void_void(PyTuple_GetItem_void_void(<void *>entry, 1), 0)
    saved_minikind = PyBytes_AsString_obj(<PyObject *>PyTuple_GetItem_void_void(details, 0))[0]
    if minikind == b'd' and saved_minikind == b't':
        minikind = b't'
    saved_link_or_sha1 = PyTuple_GetItem_void_object(details, 1)
    saved_file_size = PyTuple_GetItem_void_object(details, 2)
    saved_executable = PyTuple_GetItem_void_object(details, 3)
    saved_packed_stat = PyTuple_GetItem_void_object(details, 4)
    # Deal with pyrex decrefing the objects
    Py_INCREF(saved_link_or_sha1)
    Py_INCREF(saved_file_size)
    Py_INCREF(saved_executable)
    Py_INCREF(saved_packed_stat)
    #(saved_minikind, saved_link_or_sha1, saved_file_size,
    # saved_executable, saved_packed_stat) = entry[1][0]

    if (minikind == saved_minikind
        and packed_stat == saved_packed_stat):
        # The stat hasn't changed since we saved, so we can re-use the
        # saved sha hash.
        if minikind == b'd':
            return None

        # size should also be in packed_stat
        if saved_file_size == stat_value.st_size:
            return saved_link_or_sha1

    # If we have gotten this far, that means that we need to actually
    # process this entry.
    link_or_sha1 = None
    worth_saving = 1
    if minikind == b'f':
        executable = self._is_executable(stat_value.st_mode,
                                         saved_executable)
        if self._cutoff_time is None:
            self._sha_cutoff_time()
        if (stat_value.st_mtime < self._cutoff_time
            and stat_value.st_ctime < self._cutoff_time
            and len(entry[1]) > 1
            and entry[1][1][0] != b'a'):
                # Could check for size changes for further optimised
                # avoidance of sha1's. However the most prominent case of
                # over-shaing is during initial add, which this catches.
            link_or_sha1 = self._sha1_file(abspath)
            entry[1][0] = (b'f', link_or_sha1, stat_value.st_size,
                           executable, packed_stat)
        else:
            # This file is not worth caching the sha1. Either it is too new, or
            # it is newly added. Regardless, the only things we are changing
            # are derived from the stat, and so are not worth caching. So we do
            # *not* set the IN_MEMORY_MODIFIED flag. (But we'll save the
            # updated values if there is *other* data worth saving.)
            entry[1][0] = (b'f', b'', stat_value.st_size, executable,
                           DirState.NULLSTAT)
            worth_saving = 0
    elif minikind == b'd':
        entry[1][0] = (b'd', b'', 0, False, packed_stat)
        if saved_minikind != b'd':
            # This changed from something into a directory. Make sure we
            # have a directory block for it. This doesn't happen very
            # often, so this doesn't have to be super fast.
            block_index, entry_index, dir_present, file_present = \
                self._get_block_entry_index(entry[0][0], entry[0][1], 0)
            self._ensure_block(block_index, entry_index,
                               pathjoin(entry[0][0], entry[0][1]))
        else:
            # Any changes are derived trivially from the stat object, not worth
            # re-writing a dirstate for just this
            worth_saving = 0
    elif minikind == b'l':
        if saved_minikind == b'l':
            # If the object hasn't changed kind, it isn't worth saving the
            # dirstate just for a symlink. The default is 'fast symlinks' which
            # save the target in the inode entry, rather than separately. So to
            # stat, we've already read everything off disk.
            worth_saving = 0
        link_or_sha1 = self._read_link(abspath, saved_link_or_sha1)
        if self._cutoff_time is None:
            self._sha_cutoff_time()
        if (stat_value.st_mtime < self._cutoff_time
            and stat_value.st_ctime < self._cutoff_time):
            entry[1][0] = (b'l', link_or_sha1, stat_value.st_size,
                           False, packed_stat)
        else:
            entry[1][0] = (b'l', b'', stat_value.st_size,
                           False, DirState.NULLSTAT)
    if worth_saving:
        # Note, even though _mark_modified will only set
        # IN_MEMORY_HASH_MODIFIED, it still isn't worth 
        self._mark_modified([entry])
    return link_or_sha1


# TODO: Do we want to worry about exceptions here?
cdef char _minikind_from_string(object string) except? -1:
    """Convert a python string to a char."""
    return PyBytes_AsString(string)[0]


cdef object _kind_absent
cdef object _kind_file
cdef object _kind_directory
cdef object _kind_symlink
cdef object _kind_relocated
cdef object _kind_tree_reference
_kind_absent = "absent"
_kind_file = "file"
_kind_directory = "directory"
_kind_symlink = "symlink"
_kind_relocated = "relocated"
_kind_tree_reference = "tree-reference"


cdef object _minikind_to_kind(char minikind):
    """Create a string kind for minikind."""
    cdef char _minikind[1]
    if minikind == b'f':
        return _kind_file
    elif minikind == b'd':
        return _kind_directory
    elif minikind == b'a':
        return _kind_absent
    elif minikind == b'r':
        return _kind_relocated
    elif minikind == b'l':
        return _kind_symlink
    elif minikind == b't':
        return _kind_tree_reference
    _minikind[0] = minikind
    raise KeyError(PyBytes_FromStringAndSize(_minikind, 1))


cdef int _versioned_minikind(char minikind): # cannot_raise
    """Return non-zero if minikind is in fltd"""
    return (minikind == b'f' or
            minikind == b'd' or
            minikind == b'l' or
            minikind == b't')


cdef utf8_decode(path: bytes):
    return codecs.utf_8_decode(path, 'surrogateescape')[0]


cdef class ProcessEntryC:

    cdef int doing_consistency_expansion
    cdef object old_dirname_to_file_id # dict
    cdef object new_dirname_to_file_id # dict
    cdef object last_source_parent
    cdef object last_target_parent
    cdef int include_unchanged
    cdef int partial
    cdef object use_filesystem_for_exec
    cdef object utf8_decode
    cdef readonly object searched_specific_files
    cdef readonly object searched_exact_paths
    cdef object search_specific_files
    # The parents up to the root of the paths we are searching.
    # After all normal paths are returned, these specific items are returned.
    cdef object search_specific_file_parents
    cdef object state
    # Current iteration variables:
    cdef object current_root
    cdef object current_root_unicode
    cdef object root_entries
    cdef int root_entries_pos, root_entries_len
    cdef object root_abspath
    cdef int source_index, target_index
    cdef int want_unversioned
    cdef object tree
    cdef object dir_iterator
    cdef int block_index
    cdef object current_block
    cdef int current_block_pos
    cdef object current_block_list
    cdef object current_dir_info
    cdef object current_dir_list
    cdef object _pending_consistent_entries # list
    cdef int path_index
    cdef object root_dir_info
    cdef object bisect_left
    cdef object pathjoin
    cdef object fstat
    # A set of the ids we've output when doing partial output.
    cdef object seen_ids
    cdef object sha_file

    def __init__(self, include_unchanged, use_filesystem_for_exec,
        search_specific_files, state, source_index, target_index,
        want_unversioned, tree):
        self.doing_consistency_expansion = 0
        self.old_dirname_to_file_id = {}
        self.new_dirname_to_file_id = {}
        # Are we doing a partial iter_changes?
        self.partial = set(['']).__ne__(search_specific_files)
        # Using a list so that we can access the values and change them in
        # nested scope. Each one is [path, file_id, entry]
        self.last_source_parent = [None, None]
        self.last_target_parent = [None, None]
        if include_unchanged is None:
            self.include_unchanged = False
        else:
            self.include_unchanged = int(include_unchanged)
        self.use_filesystem_for_exec = use_filesystem_for_exec
        # for all search_indexs in each path at or under each element of
        # search_specific_files, if the detail is relocated: add the id, and
        # add the relocated path as one to search if its not searched already.
        # If the detail is not relocated, add the id.
        self.searched_specific_files = set()
        # When we search exact paths without expanding downwards, we record
        # that here.
        self.searched_exact_paths = set()
        self.search_specific_files = search_specific_files
        # The parents up to the root of the paths we are searching.
        # After all normal paths are returned, these specific items are returned.
        self.search_specific_file_parents = set()
        # The ids we've sent out in the delta.
        self.seen_ids = set()
        self.state = state
        self.current_root = None
        self.current_root_unicode = None
        self.root_entries = None
        self.root_entries_pos = 0
        self.root_entries_len = 0
        self.root_abspath = None
        if source_index is None:
            self.source_index = -1
        else:
            self.source_index = source_index
        self.target_index = target_index
        self.want_unversioned = want_unversioned
        self.tree = tree
        self.dir_iterator = None
        self.block_index = -1
        self.current_block = None
        self.current_block_list = None
        self.current_block_pos = -1
        self.current_dir_info = None
        self.current_dir_list = None
        self._pending_consistent_entries = []
        self.path_index = 0
        self.root_dir_info = None
        self.bisect_left = bisect.bisect_left
        self.pathjoin = pathjoin
        self.fstat = os.fstat
        self.sha_file = sha_file
        if target_index != 0:
            # A lot of code in here depends on target_index == 0
            raise BzrFormatsError('unsupported target index')

    cdef _process_entry(self, entry, path_info):
        """Compare an entry and real disk to generate delta information.

        :param path_info: top_relpath, basename, kind, lstat, abspath for
            the path of entry. If None, then the path is considered absent in 
            the target (Perhaps we should pass in a concrete entry for this ?)
            Basename is returned as a utf8 string because we expect this
            tuple will be ignored, and don't want to take the time to
            decode.
        :return: (iter_changes_result, changed). If the entry has not been
            handled then changed is None. Otherwise it is False if no content
            or metadata changes have occured, and True if any content or
            metadata change has occurred. If self.include_unchanged is True then
            if changed is not None, iter_changes_result will always be a result
            tuple. Otherwise, iter_changes_result is None unless changed is
            True.
        """
        cdef char target_minikind
        cdef char source_minikind
        cdef object file_id
        cdef int content_change
        cdef object details_list
        file_id = None
        details_list = entry[1]
        if -1 == self.source_index:
            source_details = DirState.NULL_PARENT_DETAILS
        else:
            source_details = details_list[self.source_index]
        target_details = details_list[self.target_index]
        target_minikind = _minikind_from_string(target_details[0])
        if path_info is not None and _versioned_minikind(target_minikind):
            if self.target_index != 0:
                raise AssertionError("Unsupported target index %d" %
                                     self.target_index)
            link_or_sha1 = update_entry(self.state, entry, path_info[4], path_info[3])
            # The entry may have been modified by update_entry
            target_details = details_list[self.target_index]
            target_minikind = _minikind_from_string(target_details[0])
        else:
            link_or_sha1 = None
        # the rest of this function is 0.3 seconds on 50K paths, or
        # 0.000006 seconds per call.
        source_minikind = _minikind_from_string(source_details[0])
        if ((_versioned_minikind(source_minikind) or source_minikind == b'r')
            and _versioned_minikind(target_minikind)):
            # claimed content in both: diff
            #   r    | fdlt   |      | add source to search, add id path move and perform
            #        |        |      | diff check on source-target
            #   r    | fdlt   |  a   | dangling file that was present in the basis.
            #        |        |      | ???
            if source_minikind != b'r':
                old_dirname = entry[0][0]
                old_basename = entry[0][1]
                old_path = path = None
            else:
                # add the source to the search path to find any children it
                # has.  TODO ? : only add if it is a container ?
                if (not self.doing_consistency_expansion and 
                    not is_inside_any(self.searched_specific_files,
                                             source_details[1])):
                    self.search_specific_files.add(source_details[1])
                    # expanding from a user requested path, parent expansion
                    # for delta consistency happens later.
                # generate the old path; this is needed for stating later
                # as well.
                old_path = source_details[1]
                old_dirname, old_basename = os.path.split(old_path)
                path = self.pathjoin(entry[0][0], entry[0][1])
                old_entry = self.state._get_entry(self.source_index,
                                             path_utf8=old_path)
                # update the source details variable to be the real
                # location.
                if old_entry == (None, None):
                    raise DirstateCorrupt(self.state._filename,
                        "entry '%s/%s' is considered renamed from %r"
                        " but source does not exist\n"
                        "entry: %s" % (entry[0][0], entry[0][1], old_path, entry))
                source_details = old_entry[1][self.source_index]
                source_minikind = _minikind_from_string(source_details[0])
            if path_info is None:
                # the file is missing on disk, show as removed.
                content_change = 1
                target_kind = None
                target_exec = False
            else:
                # source and target are both versioned and disk file is present.
                target_kind = path_info[2]
                if target_kind == 'directory':
                    if path is None:
                        old_path = path = self.pathjoin(old_dirname, old_basename)
                    file_id = entry[0][2]
                    self.new_dirname_to_file_id[path] = file_id
                    if source_minikind != b'd':
                        content_change = 1
                    else:
                        # directories have no fingerprint
                        content_change = 0
                    target_exec = False
                elif target_kind == 'file':
                    if source_minikind != b'f':
                        content_change = 1
                    else:
                        # Check the sha. We can't just rely on the size as
                        # content filtering may mean differ sizes actually
                        # map to the same content
                        if link_or_sha1 is None:
                            # Stat cache miss:
                            statvalue, link_or_sha1 = \
                                self.state._sha1_provider.stat_and_sha1(
                                path_info[4])
                            self.state._observed_sha1(entry, link_or_sha1,
                                statvalue)
                        content_change = (link_or_sha1 != source_details[1])
                    # Target details is updated at update_entry time
                    if self.use_filesystem_for_exec:
                        # We don't need S_ISREG here, because we are sure
                        # we are dealing with a file.
                        target_exec = bool(S_IXUSR & path_info[3].st_mode)
                    else:
                        target_exec = target_details[3]
                elif target_kind == 'symlink':
                    if source_minikind != b'l':
                        content_change = 1
                    else:
                        content_change = (link_or_sha1 != source_details[1])
                    target_exec = False
                elif target_kind == 'tree-reference':
                    if source_minikind != b't':
                        content_change = 1
                    else:
                        content_change = 0
                    target_exec = False
                else:
                    if path is None:
                        path = self.pathjoin(old_dirname, old_basename)
                    raise BadFileKindError(path, path_info[2])
            if source_minikind == b'd':
                if path is None:
                    old_path = path = self.pathjoin(old_dirname, old_basename)
                if file_id is None:
                    file_id = entry[0][2]
                self.old_dirname_to_file_id[old_path] = file_id
            # parent id is the entry for the path in the target tree
            if old_basename and old_dirname == self.last_source_parent[0]:
                # use a cached hit for non-root source entries.
                source_parent_id = self.last_source_parent[1]
            else:
                try:
                    source_parent_id = self.old_dirname_to_file_id[old_dirname]
                except KeyError, _:
                    source_parent_entry = self.state._get_entry(self.source_index,
                                                           path_utf8=old_dirname)
                    source_parent_id = source_parent_entry[0][2]
                if source_parent_id == entry[0][2]:
                    # This is the root, so the parent is None
                    source_parent_id = None
                else:
                    self.last_source_parent[0] = old_dirname
                    self.last_source_parent[1] = source_parent_id
            new_dirname = entry[0][0]
            if entry[0][1] and new_dirname == self.last_target_parent[0]:
                # use a cached hit for non-root target entries.
                target_parent_id = self.last_target_parent[1]
            else:
                try:
                    target_parent_id = self.new_dirname_to_file_id[new_dirname]
                except KeyError, _:
                    # TODO: We don't always need to do the lookup, because the
                    #       parent entry will be the same as the source entry.
                    target_parent_entry = self.state._get_entry(self.target_index,
                                                           path_utf8=new_dirname)
                    if target_parent_entry == (None, None):
                        raise AssertionError(
                            "Could not find target parent in wt: %s\nparent of: %s"
                            % (new_dirname, entry))
                    target_parent_id = target_parent_entry[0][2]
                if target_parent_id == entry[0][2]:
                    # This is the root, so the parent is None
                    target_parent_id = None
                else:
                    self.last_target_parent[0] = new_dirname
                    self.last_target_parent[1] = target_parent_id

            source_exec = source_details[3]
            changed = (content_change
                or source_parent_id != target_parent_id
                or old_basename != entry[0][1]
                or source_exec != target_exec
                )
            if not changed and not self.include_unchanged:
                return None, False
            else:
                if old_path is None:
                    path = self.pathjoin(old_dirname, old_basename)
                    old_path = path
                    old_path_u = utf8_decode(old_path)
                    path_u = old_path_u
                else:
                    old_path_u = utf8_decode(old_path)
                    if old_path == path:
                        path_u = old_path_u
                    else:
                        path_u = utf8_decode(path)
                source_kind = _minikind_to_kind(source_minikind)
                return DirstateInventoryChange(entry[0][2],
                       (old_path_u, path_u),
                       content_change,
                       (True, True),
                       (source_parent_id, target_parent_id),
                       (utf8_decode(old_basename), utf8_decode(entry[0][1])),
                       (source_kind, target_kind),
                       (source_exec, target_exec)), changed
        elif source_minikind == b'a' and _versioned_minikind(target_minikind):
            # looks like a new file
            path = self.pathjoin(entry[0][0], entry[0][1])
            # parent id is the entry for the path in the target tree
            # TODO: these are the same for an entire directory: cache em.
            parent_entry = self.state._get_entry(self.target_index,
                                                 path_utf8=entry[0][0])
            if parent_entry is None:
                raise DirstateCorrupt(self.state,
                    "We could not find the parent entry in index %d"
                    " for the entry: %s"
                    % (self.target_index, entry[0]))
            parent_id = parent_entry[0][2]
            if parent_id == entry[0][2]:
                parent_id = None
            if path_info is not None:
                # Present on disk:
                if self.use_filesystem_for_exec:
                    # We need S_ISREG here, because we aren't sure if this
                    # is a file or not.
                    target_exec = bool(
                        S_ISREG(path_info[3].st_mode)
                        and S_IXUSR & path_info[3].st_mode)
                else:
                    target_exec = target_details[3]
                return DirstateInventoryChange(entry[0][2],
                       (None, utf8_decode(path)),
                       True,
                       (False, True),
                       (None, parent_id),
                       (None, utf8_decode(entry[0][1])),
                       (None, path_info[2]),
                       (None, target_exec)), True
            else:
                # Its a missing file, report it as such.
                return DirstateInventoryChange(entry[0][2],
                       (None, utf8_decode(path)),
                       False,
                       (False, True),
                       (None, parent_id),
                       (None, utf8_decode(entry[0][1])),
                       (None, None),
                       (None, False)), True
        elif _versioned_minikind(source_minikind) and target_minikind == b'a':
            # unversioned, possibly, or possibly not deleted: we dont care.
            # if its still on disk, *and* theres no other entry at this
            # path [we dont know this in this routine at the moment -
            # perhaps we should change this - then it would be an unknown.
            old_path = self.pathjoin(entry[0][0], entry[0][1])
            # parent id is the entry for the path in the target tree
            parent_id = self.state._get_entry(self.source_index, path_utf8=entry[0][0])[0][2]
            if parent_id == entry[0][2]:
                parent_id = None
            return DirstateInventoryChange(
                   entry[0][2],
                   (utf8_decode(old_path), None),
                   True,
                   (True, False),
                   (parent_id, None),
                   (utf8_decode(entry[0][1]), None),
                   (_minikind_to_kind(source_minikind), None),
                   (source_details[3], None)), True
        elif _versioned_minikind(source_minikind) and target_minikind == b'r':
            # a rename; could be a true rename, or a rename inherited from
            # a renamed parent. TODO: handle this efficiently. Its not
            # common case to rename dirs though, so a correct but slow
            # implementation will do.
            if (not self.doing_consistency_expansion and 
                not is_inside_any(self.searched_specific_files,
                    target_details[1])):
                self.search_specific_files.add(target_details[1])
                # We don't expand the specific files parents list here as
                # the path is absent in target and won't create a delta with
                # missing parent.
        elif ((source_minikind == b'r' or source_minikind == b'a') and
              (target_minikind == b'r' or target_minikind == b'a')):
            # neither of the selected trees contain this path,
            # so skip over it. This is not currently directly tested, but
            # is indirectly via test_too_much.TestCommands.test_conflicts.
            pass
        else:
            raise AssertionError("don't know how to compare "
                "source_minikind=%r, target_minikind=%r"
                % (source_minikind, target_minikind))
            ## import pdb;pdb.set_trace()
        return None, None

    def __iter__(self):
        return self

    def iter_changes(self):
        return self

    cdef int _gather_result_for_consistency(self, result) except -1:
        """Check a result we will yield to make sure we are consistent later.
        
        This gathers result's parents into a set to output later.

        :param result: A result tuple.
        """
        if not self.partial or not result.file_id:
            return 0
        self.seen_ids.add(result.file_id)
        new_path = result.path[1]
        if new_path:
            # Not the root and not a delete: queue up the parents of the path.
            self.search_specific_file_parents.update(
                [p.encode('utf-8') for p in parent_directories(new_path)])
            # Add the root directory which parent_directories does not
            # provide.
            self.search_specific_file_parents.add(b'')
        return 0

    cdef int _update_current_block(self) except -1:
        if (self.block_index < len(self.state._dirblocks) and
            is_inside(self.current_root, self.state._dirblocks[self.block_index][0])):
            self.current_block = self.state._dirblocks[self.block_index]
            self.current_block_list = self.current_block[1]
            self.current_block_pos = 0
        else:
            self.current_block = None
            self.current_block_list = None
        return 0

    def __next__(self):
        # Simple thunk to allow tail recursion without pyrex confusion
        return self._iter_next()

    cdef _iter_next(self):
        """Iterate over the changes."""
        # This function single steps through an iterator. As such while loops
        # are often exited by 'return' - the code is structured so that the
        # next call into the function will return to the same while loop. Note
        # that all flow control needed to re-reach that step is reexecuted,
        # which can be a performance problem. It has not yet been tuned to
        # minimise this; a state machine is probably the simplest restructuring
        # to both minimise this overhead and make the code considerably more
        # understandable.

        # sketch: 
        # compare source_index and target_index at or under each element of search_specific_files.
        # follow the following comparison table. Note that we only want to do diff operations when
        # the target is fdl because thats when the walkdirs logic will have exposed the pathinfo 
        # for the target.
        # cases:
        # 
        # Source | Target | disk | action
        #   r    | fdlt   |      | add source to search, add id path move and perform
        #        |        |      | diff check on source-target
        #   r    | fdlt   |  a   | dangling file that was present in the basis. 
        #        |        |      | ???
        #   r    |  a     |      | add source to search
        #   r    |  a     |  a   | 
        #   r    |  r     |      | this path is present in a non-examined tree, skip.
        #   r    |  r     |  a   | this path is present in a non-examined tree, skip.
        #   a    | fdlt   |      | add new id
        #   a    | fdlt   |  a   | dangling locally added file, skip
        #   a    |  a     |      | not present in either tree, skip
        #   a    |  a     |  a   | not present in any tree, skip
        #   a    |  r     |      | not present in either tree at this path, skip as it
        #        |        |      | may not be selected by the users list of paths.
        #   a    |  r     |  a   | not present in either tree at this path, skip as it
        #        |        |      | may not be selected by the users list of paths.
        #  fdlt  | fdlt   |      | content in both: diff them
        #  fdlt  | fdlt   |  a   | deleted locally, but not unversioned - show as deleted ?
        #  fdlt  |  a     |      | unversioned: output deleted id for now
        #  fdlt  |  a     |  a   | unversioned and deleted: output deleted id
        #  fdlt  |  r     |      | relocated in this tree, so add target to search.
        #        |        |      | Dont diff, we will see an r,fd; pair when we reach
        #        |        |      | this id at the other path.
        #  fdlt  |  r     |  a   | relocated in this tree, so add target to search.
        #        |        |      | Dont diff, we will see an r,fd; pair when we reach
        #        |        |      | this id at the other path.

        # TODO: jam 20070516 - Avoid the _get_entry lookup overhead by
        #       keeping a cache of directories that we have seen.
        cdef object current_dirname, current_blockname
        cdef char * current_dirname_c
        cdef char * current_blockname_c
        cdef int advance_entry, advance_path
        cdef int path_handled
        searched_specific_files = self.searched_specific_files
        # Are we walking a root?
        while self.root_entries_pos < self.root_entries_len:
            entry = self.root_entries[self.root_entries_pos]
            self.root_entries_pos = self.root_entries_pos + 1
            result, changed = self._process_entry(entry, self.root_dir_info)
            if changed is not None:
                if changed:
                    self._gather_result_for_consistency(result)
                if changed or self.include_unchanged:
                    return result
        # Have we finished the prior root, or never started one ?
        if self.current_root is None:
            # TODO: the pending list should be lexically sorted?  the
            # interface doesn't require it.
            try:
                self.current_root = self.search_specific_files.pop()
            except KeyError, _:
                raise StopIteration()
            self.searched_specific_files.add(self.current_root)
            # process the entries for this containing directory: the rest will be
            # found by their parents recursively.
            self.root_entries = self.state._entries_for_path(self.current_root)
            self.root_entries_len = len(self.root_entries)
            self.current_root_unicode = self.current_root.decode('utf8')
            self.root_abspath = self.tree.abspath(self.current_root_unicode)
            try:
                root_stat = os.lstat(self.root_abspath)
            except OSError, e:
                if e.errno == errno.ENOENT:
                    # the path does not exist: let _process_entry know that.
                    self.root_dir_info = None
                else:
                    # some other random error: hand it up.
                    raise
            else:
                self.root_dir_info = (b'', self.current_root,
                    file_kind_from_stat_mode(root_stat.st_mode), root_stat,
                    self.root_abspath)
                if self.root_dir_info[2] == 'directory':
                    if self.tree._directory_is_tree_reference(
                        self.current_root_unicode):
                        self.root_dir_info = self.root_dir_info[:2] + \
                            ('tree-reference',) + self.root_dir_info[3:]
            if not self.root_entries and not self.root_dir_info:
                # this specified path is not present at all, skip it.
                # (tail recursion, can do a loop once the full structure is
                # known).
                return self._iter_next()
            path_handled = 0
            self.root_entries_pos = 0
            # XXX Clarity: This loop is duplicated a out the self.current_root
            # is None guard above: if we return from it, it completes there
            # (and the following if block cannot trigger because
            # path_handled must be true, so the if block is not # duplicated.
            while self.root_entries_pos < self.root_entries_len:
                entry = self.root_entries[self.root_entries_pos]
                self.root_entries_pos = self.root_entries_pos + 1
                result, changed = self._process_entry(entry, self.root_dir_info)
                if changed is not None:
                    path_handled = -1
                    if changed:
                        self._gather_result_for_consistency(result)
                    if changed or self.include_unchanged:
                        return result
            # handle unversioned specified paths:
            if self.want_unversioned and not path_handled and self.root_dir_info:
                new_executable = bool(
                    stat.S_ISREG(self.root_dir_info[3].st_mode)
                    and stat.S_IEXEC & self.root_dir_info[3].st_mode)
                return DirstateInventoryChange(
                       None,
                       (None, self.current_root_unicode),
                       True,
                       (False, False),
                       (None, None),
                       (None, splitpath(self.current_root_unicode)[-1]),
                       (None, self.root_dir_info[2]),
                       (None, new_executable)
                      )
            # If we reach here, the outer flow continues, which enters into the
            # per-root setup logic.
        if (self.current_dir_info is None and self.current_block is None and not
            self.doing_consistency_expansion):
            # setup iteration of this root:
            self.current_dir_list = None
            if self.root_dir_info and self.root_dir_info[2] == 'tree-reference':
                self.current_dir_info = None
            else:
                self.dir_iterator = _walkdirs_utf8(self.root_abspath,
                    prefix=self.current_root)
                self.path_index = 0
                try:
                    self.current_dir_info = next(self.dir_iterator)
                    self.current_dir_list = self.current_dir_info[1]
                except OSError, e:
                    # there may be directories in the inventory even though
                    # this path is not a file on disk: so mark it as end of
                    # iterator
                    if e.errno in (errno.ENOENT, errno.ENOTDIR, errno.EINVAL):
                        self.current_dir_info = None
                    elif sys.platform == 'win32':
                        # on win32, python2.4 has e.errno == ERROR_DIRECTORY, but
                        # python 2.5 has e.errno == EINVAL,
                        #            and e.winerror == ERROR_DIRECTORY
                        try:
                            e_winerror = e.winerror
                        except AttributeError, _:
                            e_winerror = None
                        win_errors = (ERROR_DIRECTORY, ERROR_PATH_NOT_FOUND)
                        if (e.errno in win_errors or e_winerror in win_errors):
                            self.current_dir_info = None
                        else:
                            # Will this really raise the right exception ?
                            raise
                    else:
                        raise
                else:
                    if self.current_dir_info[0][0] == b'':
                        # remove .bzr from iteration
                        bzr_index = self.bisect_left(self.current_dir_list, (b'.bzr',))
                        if self.current_dir_list[bzr_index][0] != b'.bzr':
                            raise AssertionError()
                        del self.current_dir_list[bzr_index]
            initial_key = (self.current_root, b'', b'')
            self.block_index, _ = self.state._find_block_index_from_key(initial_key)
            if self.block_index == 0:
                # we have processed the total root already, but because the
                # initial key matched it we should skip it here.
                self.block_index = self.block_index + 1
            self._update_current_block()
        # walk until both the directory listing and the versioned metadata
        # are exhausted. 
        while (self.current_dir_info is not None
            or self.current_block is not None):
            # Uncommon case - a missing directory or an unversioned directory:
            if (self.current_dir_info and self.current_block
                and self.current_dir_info[0][0] != self.current_block[0]):
                # Work around pyrex broken heuristic - current_dirname has
                # the same scope as current_dirname_c
                current_dirname = self.current_dir_info[0][0]
                current_dirname_c = PyBytes_AS_STRING_void(
                    <void *>current_dirname)
                current_blockname = self.current_block[0]
                current_blockname_c = PyBytes_AS_STRING_void(
                    <void *>current_blockname)
                # In the python generator we evaluate this if block once per
                # dir+block; because we reenter in the pyrex version its being
                # evaluated once per path: we could cache the result before
                # doing the while loop and probably save time.
                if _cmp_by_dirs(current_dirname_c,
                    PyBytes_Size(current_dirname),
                    current_blockname_c,
                    PyBytes_Size(current_blockname)) < 0:
                    # filesystem data refers to paths not covered by the
                    # dirblock.  this has two possibilities:
                    # A) it is versioned but empty, so there is no block for it
                    # B) it is not versioned.

                    # if (A) then we need to recurse into it to check for
                    # new unknown files or directories.
                    # if (B) then we should ignore it, because we don't
                    # recurse into unknown directories.
                    # We are doing a loop
                    while self.path_index < len(self.current_dir_list):
                        current_path_info = self.current_dir_list[self.path_index]
                        # dont descend into this unversioned path if it is
                        # a dir
                        if current_path_info[2] in ('directory',
                                                    'tree-reference'):
                            del self.current_dir_list[self.path_index]
                            self.path_index = self.path_index - 1
                        self.path_index = self.path_index + 1
                        if self.want_unversioned:
                            if current_path_info[2] == 'directory':
                                if self.tree._directory_is_tree_reference(
                                    utf8_decode(current_path_info[0])):
                                    current_path_info = current_path_info[:2] + \
                                        ('tree-reference',) + current_path_info[3:]
                            new_executable = bool(
                                stat.S_ISREG(current_path_info[3].st_mode)
                                and stat.S_IEXEC & current_path_info[3].st_mode)
                            return DirstateInventoryChange(
                                None,
                                (None, utf8_decode(current_path_info[0])),
                                True,
                                (False, False),
                                (None, None),
                                (None, utf8_decode(current_path_info[1])),
                                (None, current_path_info[2]),
                                (None, new_executable))
                    # This dir info has been handled, go to the next
                    self.path_index = 0
                    self.current_dir_list = None
                    try:
                        self.current_dir_info = next(self.dir_iterator)
                        self.current_dir_list = self.current_dir_info[1]
                    except StopIteration, _:
                        self.current_dir_info = None
                else: #(dircmp > 0)
                    # We have a dirblock entry for this location, but there
                    # is no filesystem path for this. This is most likely
                    # because a directory was removed from the disk.
                    # We don't have to report the missing directory,
                    # because that should have already been handled, but we
                    # need to handle all of the files that are contained
                    # within.
                    while self.current_block_pos < len(self.current_block_list):
                        current_entry = self.current_block_list[self.current_block_pos]
                        self.current_block_pos = self.current_block_pos + 1
                        # entry referring to file not present on disk.
                        # advance the entry only, after processing.
                        result, changed = self._process_entry(current_entry, None)
                        if changed is not None:
                            if changed:
                                self._gather_result_for_consistency(result)
                            if changed or self.include_unchanged:
                                return result
                    self.block_index = self.block_index + 1
                    self._update_current_block()
                continue # next loop-on-block/dir
            result = self._loop_one_block()
            if result is not None:
                return result
        if len(self.search_specific_files):
            # More supplied paths to process
            self.current_root = None
            return self._iter_next()
        # Start expanding more conservatively, adding paths the user may not
        # have intended but required for consistent deltas.
        self.doing_consistency_expansion = 1
        if not self._pending_consistent_entries:
            self._pending_consistent_entries = self._next_consistent_entries()
        while self._pending_consistent_entries:
            result, changed = self._pending_consistent_entries.pop()
            if changed is not None:
                return result
        raise StopIteration()

    cdef object _maybe_tree_ref(self, current_path_info):
        if self.tree._directory_is_tree_reference(
            utf8_decode(current_path_info[0])):
            return current_path_info[:2] + \
                ('tree-reference',) + current_path_info[3:]
        else:
            return current_path_info

    cdef object _loop_one_block(self):
            # current_dir_info and current_block refer to the same directory -
            # this is the common case code.
            # Assign local variables for current path and entry:
            cdef object current_entry
            cdef object current_path_info
            cdef int path_handled
            cdef char minikind
            cdef int cmp_result
            # cdef char * temp_str
            # cdef Py_ssize_t temp_str_length
            # PyBytes_AsStringAndSize(disk_kind, &temp_str, &temp_str_length)
            # if not strncmp(temp_str, "directory", temp_str_length):
            if (self.current_block is not None and
                self.current_block_pos < PyList_GET_SIZE(self.current_block_list)):
                current_entry = PyList_GET_ITEM(self.current_block_list,
                    self.current_block_pos)
                # accomodate pyrex
                Py_INCREF(current_entry)
            else:
                current_entry = None
            if (self.current_dir_info is not None and
                self.path_index < PyList_GET_SIZE(self.current_dir_list)):
                current_path_info = PyList_GET_ITEM(self.current_dir_list,
                    self.path_index)
                # accomodate pyrex
                Py_INCREF(current_path_info)
                disk_kind = PyTuple_GET_ITEM(current_path_info, 2)
                # accomodate pyrex
                Py_INCREF(disk_kind)
                if disk_kind == "directory":
                    current_path_info = self._maybe_tree_ref(current_path_info)
            else:
                current_path_info = None
            while (current_entry is not None or current_path_info is not None):
                advance_entry = -1
                advance_path = -1
                result = None
                changed = None
                path_handled = 0
                if current_entry is None:
                    # unversioned -  the check for path_handled when the path
                    # is advanced will yield this path if needed.
                    pass
                elif current_path_info is None:
                    # no path is fine: the per entry code will handle it.
                    result, changed = self._process_entry(current_entry,
                        current_path_info)
                else:
                    minikind = _minikind_from_string(
                        current_entry[1][self.target_index][0])
                    cmp_result = ((current_path_info[1] > current_entry[0][1]) -
                                  (current_path_info[1] < current_entry[0][1]))
                    if (cmp_result or minikind == b'a' or minikind == b'r'):
                        # The current path on disk doesn't match the dirblock
                        # record. Either the dirblock record is marked as
                        # absent/renamed, or the file on disk is not present at all
                        # in the dirblock. Either way, report about the dirblock
                        # entry, and let other code handle the filesystem one.

                        # Compare the basename for these files to determine
                        # which comes first
                        if cmp_result < 0:
                            # extra file on disk: pass for now, but only
                            # increment the path, not the entry
                            advance_entry = 0
                        else:
                            # entry referring to file not present on disk.
                            # advance the entry only, after processing.
                            result, changed = self._process_entry(current_entry,
                                None)
                            advance_path = 0
                    else:
                        # paths are the same,and the dirstate entry is not
                        # absent or renamed.
                        result, changed = self._process_entry(current_entry,
                            current_path_info)
                        if changed is not None:
                            path_handled = -1
                            if not changed and not self.include_unchanged:
                                changed = None
                # >- loop control starts here:
                # >- entry
                if advance_entry and current_entry is not None:
                    self.current_block_pos = self.current_block_pos + 1
                    if self.current_block_pos < PyList_GET_SIZE(self.current_block_list):
                        current_entry = self.current_block_list[self.current_block_pos]
                    else:
                        current_entry = None
                # >- path
                if advance_path and current_path_info is not None:
                    if not path_handled:
                        # unversioned in all regards
                        if self.want_unversioned:
                            new_executable = bool(
                                stat.S_ISREG(current_path_info[3].st_mode)
                                and stat.S_IEXEC & current_path_info[3].st_mode)
                            relpath_unicode = utf8_decode(current_path_info[0])
                            if changed is not None:
                                raise AssertionError(
                                    "result is not None: %r" % result)
                            result = DirstateInventoryChange(
                                None,
                                (None, relpath_unicode),
                                True,
                                (False, False),
                                (None, None),
                                (None, utf8_decode(current_path_info[1])),
                                (None, current_path_info[2]),
                                (None, new_executable))
                            changed = True
                        # dont descend into this unversioned path if it is
                        # a dir
                        if current_path_info[2] in ('directory'):
                            del self.current_dir_list[self.path_index]
                            self.path_index = self.path_index - 1
                    # dont descend the disk iterator into any tree 
                    # paths.
                    if current_path_info[2] == 'tree-reference':
                        del self.current_dir_list[self.path_index]
                        self.path_index = self.path_index - 1
                    self.path_index = self.path_index + 1
                    if self.path_index < len(self.current_dir_list):
                        current_path_info = self.current_dir_list[self.path_index]
                        if current_path_info[2] == 'directory':
                            current_path_info = self._maybe_tree_ref(
                                current_path_info)
                    else:
                        current_path_info = None
                if changed is not None:
                    # Found a result on this pass, yield it
                    if changed:
                        self._gather_result_for_consistency(result)
                    if changed or self.include_unchanged:
                        return result
            if self.current_block is not None:
                self.block_index = self.block_index + 1
                self._update_current_block()
            if self.current_dir_info is not None:
                self.path_index = 0
                self.current_dir_list = None
                try:
                    self.current_dir_info = next(self.dir_iterator)
                    self.current_dir_list = self.current_dir_info[1]
                except StopIteration, _:
                    self.current_dir_info = None

    cdef object _next_consistent_entries(self):
        """Grabs the next specific file parent case to consider.

        :return: A list of the results, each of which is as for _process_entry.
        """
        results = []
        while self.search_specific_file_parents:
            # Process the parent directories for the paths we were iterating.
            # Even in extremely large trees this should be modest, so currently
            # no attempt is made to optimise.
            path_utf8 = self.search_specific_file_parents.pop()
            if path_utf8 in self.searched_exact_paths:
                # We've examined this path.
                continue
            if is_inside_any(self.searched_specific_files, path_utf8):
                # We've examined this path.
                continue
            path_entries = self.state._entries_for_path(path_utf8)
            # We need either one or two entries. If the path in
            # self.target_index has moved (so the entry in source_index is in
            # 'ar') then we need to also look for the entry for this path in
            # self.source_index, to output the appropriate delete-or-rename.
            selected_entries = []
            found_item = False
            for candidate_entry in path_entries:
                # Find entries present in target at this path:
                if candidate_entry[1][self.target_index][0] not in (b'a', b'r'):
                    found_item = True
                    selected_entries.append(candidate_entry)
                # Find entries present in source at this path:
                elif (self.source_index is not None and
                    candidate_entry[1][self.source_index][0] not in (b'a', b'r')):
                    found_item = True
                    if candidate_entry[1][self.target_index][0] == b'a':
                        # Deleted, emit it here.
                        selected_entries.append(candidate_entry)
                    else:
                        # renamed, emit it when we process the directory it
                        # ended up at.
                        self.search_specific_file_parents.add(
                            candidate_entry[1][self.target_index][1])
            if not found_item:
                raise AssertionError(
                    "Missing entry for specific path parent %r, %r" % (
                    path_utf8, path_entries))
            path_info = self._path_info(path_utf8, path_utf8.decode('utf8'))
            for entry in selected_entries:
                if entry[0][2] in self.seen_ids:
                    continue
                result, changed = self._process_entry(entry, path_info)
                if changed is None:
                    raise AssertionError(
                        "Got entry<->path mismatch for specific path "
                        "%r entry %r path_info %r " % (
                        path_utf8, entry, path_info))
                # Only include changes - we're outside the users requested
                # expansion.
                if changed:
                    self._gather_result_for_consistency(result)
                    if (result.kind[0] == 'directory' and
                        result.kind[1] != 'directory'):
                        # This stopped being a directory, the old children have
                        # to be included.
                        if entry[1][self.source_index][0] == b'r':
                            # renamed, take the source path
                            entry_path_utf8 = entry[1][self.source_index][1]
                        else:
                            entry_path_utf8 = path_utf8
                        initial_key = (entry_path_utf8, b'', b'')
                        block_index, _ = self.state._find_block_index_from_key(
                            initial_key)
                        if block_index == 0:
                            # The children of the root are in block index 1.
                            block_index = block_index + 1
                        current_block = None
                        if block_index < len(self.state._dirblocks):
                            current_block = self.state._dirblocks[block_index]
                            if not is_inside(
                                entry_path_utf8, current_block[0]):
                                # No entries for this directory at all.
                                current_block = None
                        if current_block is not None:
                            for entry in current_block[1]:
                                if entry[1][self.source_index][0] in (b'a', b'r'):
                                    # Not in the source tree, so doesn't have to be
                                    # included.
                                    continue
                                # Path of the entry itself.
                                self.search_specific_file_parents.add(
                                    self.pathjoin(*entry[0][:2]))
                if changed or self.include_unchanged:
                    results.append((result, changed))
            self.searched_exact_paths.add(path_utf8)
        return results

    cdef object _path_info(self, utf8_path, unicode_path):
        """Generate path_info for unicode_path.

        :return: None if unicode_path does not exist, or a path_info tuple.
        """
        abspath = self.tree.abspath(unicode_path)
        try:
            stat = os.lstat(abspath)
        except OSError, e:
            if e.errno == errno.ENOENT:
                # the path does not exist.
                return None
            else:
                raise
        utf8_basename = utf8_path.rsplit(b'/', 1)[-1]
        dir_info = (utf8_path, utf8_basename,
            file_kind_from_stat_mode(stat.st_mode), stat,
            abspath)
        if dir_info[2] == 'directory':
            if self.tree._directory_is_tree_reference(
                unicode_path):
                self.root_dir_info = self.root_dir_info[:2] + \
                    ('tree-reference',) + self.root_dir_info[3:]
        return dir_info



"""Test suite for bzrformats package."""

import atexit
import difflib
import logging
import os
import re
import shutil
import sys
import tempfile
import unittest

try:
    import testtools
except ImportError:
    # Minimal compatibility if testtools is not available
    testtools = None

import importlib
from urllib.parse import quote as urlquote

from .. import osutils


def _try_import(module_name):
    """Try to import a module, returning it or None if unavailable."""
    try:
        return importlib.import_module(module_name)
    except ImportError:
        return None


def pathname2url(path):
    """Convert a local pathname to a URL path."""
    # On Unix, pathname2url is essentially identity with encoding of special chars
    # but preserving '/'
    return urlquote(path, safe="/:@")


logger = logging.getLogger("bzrformats.tests")

_unitialized_attr = object()
"""A sentinel needed to act as a default value in a method signature."""


def _rmtree_temp_dir(path, test_id=None):
    """Remove a temporary directory, handling errors."""
    try:
        shutil.rmtree(path)
    except OSError:
        if test_id:
            print(f"Failed to remove temp dir {path} for test {test_id}")
        pass


class TestCase(testtools.TestCase if testtools else unittest.TestCase):
    """Base class for bzrformats unit tests."""

    def __init__(self, methodName="testMethod"):  # noqa: N803
        super().__init__(methodName)
        self._cleanups = []

    def setUp(self):
        super().setUp()
        self._orig_cwd = os.getcwd()
        # Clear config to avoid external config affecting tests
        # Override HOME to prevent reading user configs
        import tempfile

        self._test_home_dir = tempfile.mkdtemp(prefix="brz-test-home-")
        self.addCleanup(__import__("shutil").rmtree, self._test_home_dir)
        self.overrideEnv("HOME", self._test_home_dir)
        self.overrideEnv("BRZ_HOME", self._test_home_dir)
        self.overrideEnv("EMAIL", "jrandom@example.com")
        self.overrideEnv("BRZ_EMAIL", None)

    def tearDown(self):
        try:
            # Run any registered cleanup functions
            while self._cleanups:
                func, args, kwargs = self._cleanups.pop()
                func(*args, **kwargs)
        finally:
            os.chdir(self._orig_cwd)
            super().tearDown()

    def addCleanup(self, func, *args, **kwargs):
        """Register a function to be called during tearDown."""
        self._cleanups.append((func, args, kwargs))

    def overrideAttr(self, obj, attr_name, new=_unitialized_attr):
        """Overrides an object attribute restoring it after the test."""
        # The actual value is captured by the call below
        value = getattr(obj, attr_name, _unitialized_attr)
        if value is _unitialized_attr:
            # When the test completes, the attribute should not exist, but if
            # we aren't setting a value, we don't need to do anything.
            if new is not _unitialized_attr:
                self.addCleanup(delattr, obj, attr_name)
        else:
            self.addCleanup(setattr, obj, attr_name, value)
        if new is not _unitialized_attr:
            setattr(obj, attr_name, new)
        return value

    def overrideEnv(self, name, new_value):
        """Override an environment variable, restoring it during tearDown."""
        old_value = os.environ.get(name)
        if new_value is None:
            if name in os.environ:
                del os.environ[name]
        else:
            os.environ[name] = new_value

        def restore():
            if old_value is None:
                if name in os.environ:
                    del os.environ[name]
            else:
                os.environ[name] = old_value

        self.addCleanup(restore)

    def assertEqualDiff(self, a, b, message=None):
        """Assert two texts are equal, if not raise an exception showing diffs."""
        if a == b:
            return
        if message is None:
            message = "texts not equal:\n"
        if a + "\n" == b:
            message = "first string is missing a final newline.\n"
        if a == b + "\n":
            message = "second string is missing a final newline.\n"

        # Create a diff
        diff = difflib.unified_diff(
            a.splitlines(True), b.splitlines(True), "expected", "actual"
        )
        raise AssertionError(message + "".join(diff))

    def assertContainsRe(self, haystack, needle_re, flags=0):
        """Assert that haystack contains something matching a regular expression."""
        if not re.search(needle_re, haystack, flags):
            raise AssertionError(f'pattern "{needle_re}" not found in "{haystack}"')

    def assertNotContainsRe(self, haystack, needle_re, flags=0):
        """Assert that haystack does not match a regular expression."""
        if re.search(needle_re, haystack, flags):
            raise AssertionError(f'pattern "{needle_re}" found in "{haystack}"')

    def assertStartsWith(self, s, prefix):
        if not s.startswith(prefix):
            raise AssertionError(f"string {s!r} does not start with {prefix!r}")

    def assertEndsWith(self, s, suffix):
        if not s.endswith(suffix):
            raise AssertionError(f"string {s!r} does not end with {suffix!r}")

    def assertLength(self, expected_length, obj_with_len):
        """Assert that obj_with_len is of length expected_length."""
        actual_length = len(obj_with_len)
        if actual_length != expected_length:
            self.fail(
                f"Incorrect length: wanted {expected_length}, got {actual_length} for {obj_with_len!r}"
            )

    def assertIs(self, left, right, message=None):
        """Assert that left is right."""
        if left is not right:
            if message is not None:
                raise AssertionError(message)
            else:
                raise AssertionError(f"{left!r} is not {right!r}.")

    def assertIsNot(self, left, right, message=None):
        """Assert that left is not right."""
        if left is right:
            if message is not None:
                raise AssertionError(message)
            else:
                raise AssertionError(f"{left!r} is {right!r}.")

    def assertIsInstance(self, obj, klass, msg=None):
        """Assert that obj is an instance of klass."""
        if not isinstance(obj, klass):
            if msg is None:
                msg = f"{obj!r} is not an instance of {klass}"
            raise AssertionError(msg)

    def log(self, *args):
        """Log a message."""
        logger.debug(*args)

    def assertSubset(self, sublist, superlist):
        """Assert that every entry in sublist is present in superlist."""
        missing = set(sublist) - set(superlist)
        if missing:
            raise AssertionError(
                f"Missing elements {missing!r}: {sublist!r} not a subset of {superlist!r}"
            )

    def knownFailure(self, reason):
        """Mark test as a known failure."""
        raise expectedFailure(reason)

    def requireFeature(self, feature):
        """This test requires a specific feature is available.

        :raises unittest.SkipTest: When feature is not available.
        """
        if not feature.available():
            self.skipTest(f"Feature {feature.feature_name()} not available")

    def assertPathExists(self, path):
        """Fail unless path or paths, which may be abs or relative, exist."""
        if not isinstance(path, (bytes, str)):
            for p in path:
                if not os.path.exists(p):
                    self.fail(f"path {p} does not exist")
        else:
            if not os.path.exists(path):
                self.fail(f"path {path} does not exist")

    def assertPathDoesNotExist(self, path):
        """Fail if path or paths, which may be abs or relative, exist."""
        if not isinstance(path, (bytes, str)):
            for p in path:
                if os.path.exists(p):
                    self.fail(f"path {p} exists")
        else:
            if os.path.exists(path):
                self.fail(f"path {path} exists")

    def assertFileEqual(self, content, path):
        """Fail if path does not contain 'content'."""
        self.assertPathExists(path)

        mode = "r" + ("b" if isinstance(content, bytes) else "")
        with open(path, mode) as f:
            s = f.read()
        self.assertEqualDiff(content, s)

    def assertListRaises(self, excClass, func, *args, **kwargs):  # noqa: N803
        """Fail unless excClass is raised when the iterator from func is used.

        Many functions can return generators this makes sure
        to wrap them in a list() call to make sure the whole generator
        is run, and that the proper exception is raised.
        """
        try:
            list(func(*args, **kwargs))
        except excClass as e:
            return e
        else:
            if getattr(excClass, "__name__", None) is not None:
                excName = excClass.__name__
            else:
                excName = str(excClass)
            raise self.failureException(f"{excName} not raised")

    def time(self, callable, *args, **kwargs):
        """Run callable and return result."""
        # Simplified version - just run the callable without profiling
        return callable(*args, **kwargs)


class TestCaseInTempDir(TestCase):
    """Test case that runs in a temporary directory.

    This is a minimal version of brz's TestCaseInTempDir.
    """

    TEST_ROOT = None

    def setUp(self):
        super().setUp()
        self._make_test_root()
        self.addCleanup(os.chdir, os.getcwd())
        self.makeAndChdirToTestDir()

    def _make_test_root(self):
        """Create the top-level test directory if needed."""
        if TestCaseInTempDir.TEST_ROOT is None:
            root = os.path.realpath(
                tempfile.mkdtemp(prefix="testbzrformats-", suffix=".tmp")
            )
            TestCaseInTempDir.TEST_ROOT = root
            atexit.register(_rmtree_temp_dir, root)

    def makeAndChdirToTestDir(self):
        """Create a temporary directory for this test and chdir to it."""
        # Create test directory name based on test id
        test_name = self.id()
        if sys.platform in ("win32", "cygwin"):
            test_name = re.sub('[<>*=+",:;_/\\-]', "_", test_name)
            test_name = test_name[-30:]  # Windows path length limits
        else:
            test_name = re.sub("[/]", "_", test_name)

        base_dir = os.path.join(TestCaseInTempDir.TEST_ROOT, test_name)

        # Find a unique directory name
        test_dir = base_dir
        for i in range(100):
            if not os.path.exists(test_dir):
                break
            test_dir = f"{base_dir}_{i}"
        else:
            raise RuntimeError(
                f"Could not create unique test directory for {test_name}"
            )

        os.makedirs(test_dir)
        self.test_dir = test_dir
        self.addCleanup(_rmtree_temp_dir, test_dir, test_id=self.id())
        os.chdir(test_dir)

    def build_tree(self, shape, line_endings="binary", transport=None):
        """Build a test tree according to a pattern.

        shape is a sequence of file specifications. If the final
        character is '/', a directory is created.
        """
        for name in shape:
            if isinstance(name, tuple):
                name, content = name
            else:
                content = None

            if name.endswith("/"):
                os.makedirs(name, exist_ok=True)
            else:
                dirname = os.path.dirname(name)
                if dirname:
                    os.makedirs(dirname, exist_ok=True)
                if content is None:
                    content = f"contents of {name}\n"
                if isinstance(content, str):
                    if line_endings == "native":
                        content = content.replace("\n", os.linesep)
                    content = content.encode("utf-8")
                with open(name, "wb") as f:
                    f.write(content)

    @staticmethod
    def build_tree_contents(shape):
        """Build test files with specific contents."""
        for entry in shape:
            if len(entry) == 2:
                name, content = entry
            else:
                name = entry[0]
                content = None
            if name.endswith("/"):
                os.makedirs(name, exist_ok=True)
            else:
                dirname = os.path.dirname(name)
                if dirname:
                    os.makedirs(dirname, exist_ok=True)
                if content is None:
                    content = b""
                if isinstance(content, str):
                    content = content.encode("utf-8")
                with open(name, "wb") as f:
                    f.write(content)


# Import TestSkipped from unittest
TestSkipped = unittest.SkipTest


class TestNotApplicable(TestSkipped):
    """Skip a test because it is not applicable to the current configuration."""

    pass


class TestCaseWithMemoryTransport(TestCase):
    """TestCase with a MemoryTransport for testing.

    Uses bzrformats' own MemoryTransport. Each test gets a fresh
    transport namespace based on the test ID.
    """

    def setUp(self):
        super().setUp()
        from ..transport import MemoryTransport

        self._memory_transport = MemoryTransport(url=f"memory:///{self.id()}/")

    def get_transport(self, relpath=None):
        """Get the transport for this test case."""
        if relpath is None or relpath == ".":
            return self._memory_transport
        t = self._memory_transport.clone(relpath)
        t.ensure_base()
        return t

    def get_url(self, relpath=None):
        """Get a URL for the memory transport."""
        if relpath is None or relpath == ".":
            return self._memory_transport.base
        return self._memory_transport.abspath(relpath)

    def check_file_contents(self, filename, expect):
        """Check contents of a file on the transport."""
        contents = self.get_transport().get_bytes(filename)
        if contents != expect:
            self.log(f"expected: {expect!r}")
            self.log(f"actually: {contents!r}")
            self.fail(f"contents of {filename} not as expected")


def load_tests(loader, basic_tests, pattern):
    """Load tests for bzrformats using the standard unittest discovery mechanism."""
    suite = loader.suiteClass()
    # Add the tests for this module
    suite.addTests(basic_tests)

    # List of test modules to load
    testmod_names = [
        "per_inventory",
        "per_versionedfile",
        "test__btree_serializer",
        "test__chk_map",
        "test__dirstate_helpers",
        "test__groupcompress",
        "test_btree_index",
        "test_chk_map",
        "test_chk_serializer",
        "test_chunk_writer",
        "test_dirstate",
        "test_generate_ids",
        "test_groupcompress",
        "test_hashcache",
        "test_index",
        "test_inv",
        "test_inventory_delta",
        "test_knit",
        "test_pack",
        "test_rio",
        "test_serializer",
        "test_tuned_gzip",
        "test_versionedfile",
        "test_weave",
        "test_xml",
    ]

    # Load each test module
    prefix = __name__ + "."
    for testmod_name in testmod_names:
        suite.addTest(loader.loadTestsFromName(prefix + testmod_name))

    # Also load per_* modules
    per_modules = [
        "per_versionedfile",
        "per_inventory",
    ]

    for per_module in per_modules:
        try:
            suite.addTest(loader.loadTestsFromName(prefix + per_module))
        except (ImportError, AttributeError):
            # Skip if module doesn't exist or has no tests
            pass

    return suite


def test_suite():
    """Return the test suite for bzrformats (for backwards compatibility)."""
    loader = unittest.TestLoader()
    basic_tests = loader.loadTestsFromModule(__import__(__name__, fromlist=[""]))
    return load_tests(loader, basic_tests, None)


def dir_reader_scenarios():
    """Simplified dir_reader_scenarios for bzrformats tests."""
    # Only use the unicode reader which is always available
    return [
        (
            "unicode",
            {
                "_dir_reader_class": osutils.UnicodeDirReader,
                "_native_to_unicode": lambda x: x,  # Already unicode
            },
        )
    ]



# Copyright (C) 2006-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests of the dirstate functionality being built for WorkingTreeFormat4."""

import binascii
import bisect
import os
import struct

from testscenarios import load_tests_apply_scenarios

from bzrformats import osutils

from .. import dirstate, inventory
from . import TestCase, TestCaseInTempDir, dir_reader_scenarios

# TODO:
# TESTS to write:
# general checks for NOT_IN_MEMORY error conditions.
# set_path_id on a NOT_IN_MEMORY dirstate
# set_path_id  unicode support
# set_path_id  setting id of a path not root
# set_path_id  setting id when there are parents without the id in the parents
# set_path_id  setting id when there are parents with the id in the parents
# set_path_id  setting id when state is not in memory
# set_path_id  setting id when state is in memory unmodified
# set_path_id  setting id when state is in memory modified


class TestErrors(TestCase):
    def test_dirstate_corrupt(self):
        error = dirstate.DirstateCorrupt(
            ".bzr/checkout/dirstate", 'trailing garbage: "x"'
        )
        self.assertEqualDiff(
            "The dirstate file (.bzr/checkout/dirstate)"
            ' appears to be corrupt: trailing garbage: "x"',
            str(error),
        )


load_tests = load_tests_apply_scenarios


class TestCaseWithDirState:
    """Helper methods for creating DirState objects.

    Inherit from this alongside a TestCase that provides a temp directory.
    """

    scenarios = dir_reader_scenarios()

    # Set by load_tests
    _dir_reader_class = None
    _native_to_unicode = None  # Not used yet

    def setUp(self):
        super().setUp()
        if self._dir_reader_class is None:
            self._dir_reader_class = osutils.UnicodeDirReader
        self.overrideAttr(osutils, "_selected_dir_reader", self._dir_reader_class())

    def create_empty_dirstate(self):
        """Return a locked but empty dirstate."""
        state = dirstate.DirState.initialize("dirstate")
        return state

    def create_dirstate_with_root(self):
        """Return a write-locked state with a single root entry."""
        packed_stat = b"AAAAREUHaIpFB2iKAAADAQAtkqUAAIGk"
        root_entry_direntry = (
            (b"", b"", b"a-root-value"),
            [
                (b"d", b"", 0, False, packed_stat),
            ],
        )
        dirblocks = []
        dirblocks.append((b"", [root_entry_direntry]))
        dirblocks.append((b"", []))
        state = self.create_empty_dirstate()
        try:
            state._set_data([], dirblocks)
            state._validate()
        except:
            state.unlock()
            raise
        return state

    def create_dirstate_with_root_and_subdir(self):
        """Return a locked DirState with a root and a subdir."""
        packed_stat = b"AAAAREUHaIpFB2iKAAADAQAtkqUAAIGk"
        subdir_entry = (
            (b"", b"subdir", b"subdir-id"),
            [
                (b"d", b"", 0, False, packed_stat),
            ],
        )
        state = self.create_dirstate_with_root()
        try:
            dirblocks = list(state._dirblocks)
            dirblocks[1][1].append(subdir_entry)
            state._set_data([], dirblocks)
        except:
            state.unlock()
            raise
        return state

    def create_complex_dirstate(self):
        r"""This dirstate contains multiple files and directories.

         /        a-root-value
         a/       a-dir
         b/       b-dir
         c        c-file
         d        d-file
         a/e/     e-dir
         a/f      f-file
         b/g      g-file
         b/h\xc3\xa5  h-\xc3\xa5-file  #This is u'\xe5' encoded into utf-8

        Notice that a/e is an empty directory.

        :return: The dirstate, still write-locked.
        """
        packed_stat = b"AAAAREUHaIpFB2iKAAADAQAtkqUAAIGk"
        null_sha = b"xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"
        root_entry = (
            (b"", b"", b"a-root-value"),
            [
                (b"d", b"", 0, False, packed_stat),
            ],
        )
        a_entry = (
            (b"", b"a", b"a-dir"),
            [
                (b"d", b"", 0, False, packed_stat),
            ],
        )
        b_entry = (
            (b"", b"b", b"b-dir"),
            [
                (b"d", b"", 0, False, packed_stat),
            ],
        )
        c_entry = (
            (b"", b"c", b"c-file"),
            [
                (b"f", null_sha, 10, False, packed_stat),
            ],
        )
        d_entry = (
            (b"", b"d", b"d-file"),
            [
                (b"f", null_sha, 20, False, packed_stat),
            ],
        )
        e_entry = (
            (b"a", b"e", b"e-dir"),
            [
                (b"d", b"", 0, False, packed_stat),
            ],
        )
        f_entry = (
            (b"a", b"f", b"f-file"),
            [
                (b"f", null_sha, 30, False, packed_stat),
            ],
        )
        g_entry = (
            (b"b", b"g", b"g-file"),
            [
                (b"f", null_sha, 30, False, packed_stat),
            ],
        )
        h_entry = (
            (b"b", b"h\xc3\xa5", b"h-\xc3\xa5-file"),
            [
                (b"f", null_sha, 40, False, packed_stat),
            ],
        )
        dirblocks = []
        dirblocks.append((b"", [root_entry]))
        dirblocks.append((b"", [a_entry, b_entry, c_entry, d_entry]))
        dirblocks.append((b"a", [e_entry, f_entry]))
        dirblocks.append((b"b", [g_entry, h_entry]))
        state = dirstate.DirState.initialize("dirstate")
        state._validate()
        try:
            state._set_data([], dirblocks)
        except:
            state.unlock()
            raise
        return state

    def check_state_with_reopen(self, expected_result, state):
        """Check that state has current state expected_result.

        This will check the current state, open the file anew and check it
        again.
        This function expects the current state to be locked for writing, and
        will unlock it before re-opening.
        This is required because we can't open a lock_read() while something
        else has a lock_write().
            write => mutually exclusive lock
            read => shared lock
        """
        # The state should already be write locked, since we just had to do
        # some operation to get here.
        self.assertIsNotNone(state._lock_token)
        try:
            self.assertEqual(expected_result[0], state.get_parent_ids())
            # there should be no ghosts in this tree.
            self.assertEqual([], state.get_ghosts())
            # there should be one fileid in this tree - the root of the tree.
            self.assertEqual(expected_result[1], list(state._iter_entries()))
            state.save()
        finally:
            state.unlock()
        del state
        state = dirstate.DirState.on_file("dirstate")
        state.lock_read()
        try:
            self.assertEqual(expected_result[1], list(state._iter_entries()))
        finally:
            state.unlock()


class TestDirStateInitialize(TestCaseWithDirState, TestCaseInTempDir):
    def test_initialize(self):
        expected_result = (
            [],
            [
                (
                    (b"", b"", b"TREE_ROOT"),  # common details
                    [
                        (
                            b"d",
                            b"",
                            0,
                            False,
                            dirstate.DirState.NULLSTAT,
                        ),  # current tree
                    ],
                )
            ],
        )
        state = dirstate.DirState.initialize("dirstate")
        try:
            self.assertIsInstance(state, dirstate.DirState)
            lines = state.get_lines()
        finally:
            state.unlock()
        # On win32 you can't read from a locked file, even within the same
        # process. So we have to unlock and release before we check the file
        # contents.
        self.assertFileEqual(b"".join(lines), "dirstate")
        state.lock_read()  # check_state_with_reopen will unlock
        self.check_state_with_reopen(expected_result, state)


class TestGetLines(TestCaseWithDirState, TestCaseInTempDir):
    def test_get_line_with_2_rows(self):
        state = self.create_dirstate_with_root_and_subdir()
        try:
            self.assertEqual(
                [
                    b"#bazaar dirstate flat format 3\n",
                    b"crc32: 41262208\n",
                    b"num_entries: 2\n",
                    b"0\x00\n\x00"
                    b"0\x00\n\x00"
                    b"\x00\x00a-root-value\x00"
                    b"d\x00\x000\x00n\x00AAAAREUHaIpFB2iKAAADAQAtkqUAAIGk\x00\n\x00"
                    b"\x00subdir\x00subdir-id\x00"
                    b"d\x00\x000\x00n\x00AAAAREUHaIpFB2iKAAADAQAtkqUAAIGk\x00\n\x00",
                ],
                state.get_lines(),
            )
        finally:
            state.unlock()

    def test_entry_to_line(self):
        state = self.create_dirstate_with_root()
        try:
            self.assertEqual(
                b"\x00\x00a-root-value\x00d\x00\x000\x00n"
                b"\x00AAAAREUHaIpFB2iKAAADAQAtkqUAAIGk",
                state._entry_to_line(state._dirblocks[0][1][0]),
            )
        finally:
            state.unlock()

    def test_entry_to_line_with_parent(self):
        packed_stat = b"AAAAREUHaIpFB2iKAAADAQAtkqUAAIGk"
        root_entry = (
            (b"", b"", b"a-root-value"),
            [
                (b"d", b"", 0, False, packed_stat),  # current tree details
                # first: a pointer to the current location
                (b"a", b"dirname/basename", 0, False, b""),
            ],
        )
        state = dirstate.DirState.initialize("dirstate")
        try:
            self.assertEqual(
                b"\x00\x00a-root-value\x00"
                b"d\x00\x000\x00n\x00AAAAREUHaIpFB2iKAAADAQAtkqUAAIGk\x00"
                b"a\x00dirname/basename\x000\x00n\x00",
                state._entry_to_line(root_entry),
            )
        finally:
            state.unlock()

    def test_entry_to_line_with_two_parents_at_different_paths(self):
        # / in the tree, at / in one parent and /dirname/basename in the other.
        packed_stat = b"AAAAREUHaIpFB2iKAAADAQAtkqUAAIGk"
        root_entry = (
            (b"", b"", b"a-root-value"),
            [
                (b"d", b"", 0, False, packed_stat),  # current tree details
                (b"d", b"", 0, False, b"rev_id"),  # first parent details
                # second: a pointer to the current location
                (b"a", b"dirname/basename", 0, False, b""),
            ],
        )
        state = dirstate.DirState.initialize("dirstate")
        try:
            self.assertEqual(
                b"\x00\x00a-root-value\x00"
                b"d\x00\x000\x00n\x00AAAAREUHaIpFB2iKAAADAQAtkqUAAIGk\x00"
                b"d\x00\x000\x00n\x00rev_id\x00"
                b"a\x00dirname/basename\x000\x00n\x00",
                state._entry_to_line(root_entry),
            )
        finally:
            state.unlock()

    def test_iter_entries(self):
        # we should be able to iterate the dirstate entries from end to end
        # this is for get_lines to be easy to read.
        packed_stat = b"AAAAREUHaIpFB2iKAAADAQAtkqUAAIGk"
        dirblocks = []
        root_entries = [
            (
                (b"", b"", b"a-root-value"),
                [
                    (b"d", b"", 0, False, packed_stat),  # current tree details
                ],
            )
        ]
        dirblocks.append(("", root_entries))
        # add two files in the root
        subdir_entry = (
            (b"", b"subdir", b"subdir-id"),
            [
                (b"d", b"", 0, False, packed_stat),  # current tree details
            ],
        )
        afile_entry = (
            (b"", b"afile", b"afile-id"),
            [
                (b"f", b"sha1value", 34, False, packed_stat),  # current tree details
            ],
        )
        dirblocks.append(("", [subdir_entry, afile_entry]))
        # and one in subdir
        file_entry2 = (
            (b"subdir", b"2file", b"2file-id"),
            [
                (b"f", b"sha1value", 23, False, packed_stat),  # current tree details
            ],
        )
        dirblocks.append(("subdir", [file_entry2]))
        state = dirstate.DirState.initialize("dirstate")
        try:
            state._set_data([], dirblocks)
            expected_entries = [root_entries[0], subdir_entry, afile_entry, file_entry2]
            self.assertEqual(expected_entries, list(state._iter_entries()))
        finally:
            state.unlock()


class TestGetBlockRowIndex(TestCaseWithDirState, TestCaseInTempDir):
    def assertBlockRowIndexEqual(
        self,
        block_index,
        row_index,
        dir_present,
        file_present,
        state,
        dirname,
        basename,
        tree_index,
    ):
        self.assertEqual(
            (block_index, row_index, dir_present, file_present),
            state._get_block_entry_index(dirname, basename, tree_index),
        )
        if dir_present:
            block = state._dirblocks[block_index]
            self.assertEqual(dirname, block[0])
        if dir_present and file_present:
            row = state._dirblocks[block_index][1][row_index]
            self.assertEqual(dirname, row[0][0])
            self.assertEqual(basename, row[0][1])

    def test_simple_structure(self):
        state = self.create_dirstate_with_root_and_subdir()
        self.addCleanup(state.unlock)
        self.assertBlockRowIndexEqual(1, 0, True, True, state, b"", b"subdir", 0)
        self.assertBlockRowIndexEqual(1, 0, True, False, state, b"", b"bdir", 0)
        self.assertBlockRowIndexEqual(1, 1, True, False, state, b"", b"zdir", 0)
        self.assertBlockRowIndexEqual(2, 0, False, False, state, b"a", b"foo", 0)
        self.assertBlockRowIndexEqual(2, 0, False, False, state, b"subdir", b"foo", 0)

    def test_complex_structure_exists(self):
        state = self.create_complex_dirstate()
        self.addCleanup(state.unlock)
        # Make sure we can find everything that exists
        self.assertBlockRowIndexEqual(0, 0, True, True, state, b"", b"", 0)
        self.assertBlockRowIndexEqual(1, 0, True, True, state, b"", b"a", 0)
        self.assertBlockRowIndexEqual(1, 1, True, True, state, b"", b"b", 0)
        self.assertBlockRowIndexEqual(1, 2, True, True, state, b"", b"c", 0)
        self.assertBlockRowIndexEqual(1, 3, True, True, state, b"", b"d", 0)
        self.assertBlockRowIndexEqual(2, 0, True, True, state, b"a", b"e", 0)
        self.assertBlockRowIndexEqual(2, 1, True, True, state, b"a", b"f", 0)
        self.assertBlockRowIndexEqual(3, 0, True, True, state, b"b", b"g", 0)
        self.assertBlockRowIndexEqual(3, 1, True, True, state, b"b", b"h\xc3\xa5", 0)

    def test_complex_structure_missing(self):
        state = self.create_complex_dirstate()
        self.addCleanup(state.unlock)
        # Make sure things would be inserted in the right locations
        # '_' comes before 'a'
        self.assertBlockRowIndexEqual(0, 0, True, True, state, b"", b"", 0)
        self.assertBlockRowIndexEqual(1, 0, True, False, state, b"", b"_", 0)
        self.assertBlockRowIndexEqual(1, 1, True, False, state, b"", b"aa", 0)
        self.assertBlockRowIndexEqual(1, 4, True, False, state, b"", b"h\xc3\xa5", 0)
        self.assertBlockRowIndexEqual(2, 0, False, False, state, b"_", b"a", 0)
        self.assertBlockRowIndexEqual(3, 0, False, False, state, b"aa", b"a", 0)
        self.assertBlockRowIndexEqual(4, 0, False, False, state, b"bb", b"a", 0)
        # This would be inserted between a/ and b/
        self.assertBlockRowIndexEqual(3, 0, False, False, state, b"a/e", b"a", 0)
        # Put at the end
        self.assertBlockRowIndexEqual(4, 0, False, False, state, b"e", b"a", 0)


class TestGetEntry(TestCaseWithDirState, TestCaseInTempDir):
    def assertEntryEqual(self, dirname, basename, file_id, state, path, index):
        """Check that the right entry is returned for a request to getEntry."""
        entry = state._get_entry(index, path_utf8=path)
        if file_id is None:
            self.assertEqual((None, None), entry)
        else:
            cur = entry[0]
            self.assertEqual((dirname, basename, file_id), cur[:3])

    def test_simple_structure(self):
        state = self.create_dirstate_with_root_and_subdir()
        self.addCleanup(state.unlock)
        self.assertEntryEqual(b"", b"", b"a-root-value", state, b"", 0)
        self.assertEntryEqual(b"", b"subdir", b"subdir-id", state, b"subdir", 0)
        self.assertEntryEqual(None, None, None, state, b"missing", 0)
        self.assertEntryEqual(None, None, None, state, b"missing/foo", 0)
        self.assertEntryEqual(None, None, None, state, b"subdir/foo", 0)

    def test_complex_structure_exists(self):
        state = self.create_complex_dirstate()
        self.addCleanup(state.unlock)
        self.assertEntryEqual(b"", b"", b"a-root-value", state, b"", 0)
        self.assertEntryEqual(b"", b"a", b"a-dir", state, b"a", 0)
        self.assertEntryEqual(b"", b"b", b"b-dir", state, b"b", 0)
        self.assertEntryEqual(b"", b"c", b"c-file", state, b"c", 0)
        self.assertEntryEqual(b"", b"d", b"d-file", state, b"d", 0)
        self.assertEntryEqual(b"a", b"e", b"e-dir", state, b"a/e", 0)
        self.assertEntryEqual(b"a", b"f", b"f-file", state, b"a/f", 0)
        self.assertEntryEqual(b"b", b"g", b"g-file", state, b"b/g", 0)
        self.assertEntryEqual(
            b"b", b"h\xc3\xa5", b"h-\xc3\xa5-file", state, b"b/h\xc3\xa5", 0
        )

    def test_complex_structure_missing(self):
        state = self.create_complex_dirstate()
        self.addCleanup(state.unlock)
        self.assertEntryEqual(None, None, None, state, b"_", 0)
        self.assertEntryEqual(None, None, None, state, b"_\xc3\xa5", 0)
        self.assertEntryEqual(None, None, None, state, b"a/b", 0)
        self.assertEntryEqual(None, None, None, state, b"c/d", 0)

    def test_get_entry_uninitialized(self):
        """Calling get_entry will load data if it needs to."""
        state = self.create_dirstate_with_root()
        try:
            state.save()
        finally:
            state.unlock()
        del state
        state = dirstate.DirState.on_file("dirstate")
        state.lock_read()
        try:
            self.assertEqual(dirstate.DirState.NOT_IN_MEMORY, state._header_state)
            self.assertEqual(dirstate.DirState.NOT_IN_MEMORY, state._dirblock_state)
            self.assertEntryEqual(b"", b"", b"a-root-value", state, b"", 0)
        finally:
            state.unlock()


class TestIterChildEntries(TestCaseWithDirState, TestCaseInTempDir):
    def create_dirstate_with_two_trees(self):
        r"""This dirstate contains multiple files and directories.

         /        a-root-value
         a/       a-dir
         b/       b-dir
         c        c-file
         d        d-file
         a/e/     e-dir
         a/f      f-file
         b/g      g-file
         b/h\xc3\xa5  h-\xc3\xa5-file  #This is u'\xe5' encoded into utf-8

        Notice that a/e is an empty directory.

        There is one parent tree, which has the same shape with the following variations:
        b/g in the parent is gone.
        b/h in the parent has a different id
        b/i is new in the parent
        c is renamed to b/j in the parent

        :return: The dirstate, still write-locked.
        """
        packed_stat = b"AAAAREUHaIpFB2iKAAADAQAtkqUAAIGk"
        null_sha = b"xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"
        NULL_PARENT_DETAILS = dirstate.DirState.NULL_PARENT_DETAILS
        root_entry = (
            (b"", b"", b"a-root-value"),
            [
                (b"d", b"", 0, False, packed_stat),
                (b"d", b"", 0, False, b"parent-revid"),
            ],
        )
        a_entry = (
            (b"", b"a", b"a-dir"),
            [
                (b"d", b"", 0, False, packed_stat),
                (b"d", b"", 0, False, b"parent-revid"),
            ],
        )
        b_entry = (
            (b"", b"b", b"b-dir"),
            [
                (b"d", b"", 0, False, packed_stat),
                (b"d", b"", 0, False, b"parent-revid"),
            ],
        )
        c_entry = (
            (b"", b"c", b"c-file"),
            [
                (b"f", null_sha, 10, False, packed_stat),
                (b"r", b"b/j", 0, False, b""),
            ],
        )
        d_entry = (
            (b"", b"d", b"d-file"),
            [
                (b"f", null_sha, 20, False, packed_stat),
                (b"f", b"d", 20, False, b"parent-revid"),
            ],
        )
        e_entry = (
            (b"a", b"e", b"e-dir"),
            [
                (b"d", b"", 0, False, packed_stat),
                (b"d", b"", 0, False, b"parent-revid"),
            ],
        )
        f_entry = (
            (b"a", b"f", b"f-file"),
            [
                (b"f", null_sha, 30, False, packed_stat),
                (b"f", b"f", 20, False, b"parent-revid"),
            ],
        )
        g_entry = (
            (b"b", b"g", b"g-file"),
            [
                (b"f", null_sha, 30, False, packed_stat),
                NULL_PARENT_DETAILS,
            ],
        )
        h_entry1 = (
            (b"b", b"h\xc3\xa5", b"h-\xc3\xa5-file1"),
            [
                (b"f", null_sha, 40, False, packed_stat),
                NULL_PARENT_DETAILS,
            ],
        )
        h_entry2 = (
            (b"b", b"h\xc3\xa5", b"h-\xc3\xa5-file2"),
            [
                NULL_PARENT_DETAILS,
                (b"f", b"h", 20, False, b"parent-revid"),
            ],
        )
        i_entry = (
            (b"b", b"i", b"i-file"),
            [
                NULL_PARENT_DETAILS,
                (b"f", b"h", 20, False, b"parent-revid"),
            ],
        )
        j_entry = (
            (b"b", b"j", b"c-file"),
            [
                (b"r", b"c", 0, False, b""),
                (b"f", b"j", 20, False, b"parent-revid"),
            ],
        )
        dirblocks = []
        dirblocks.append((b"", [root_entry]))
        dirblocks.append((b"", [a_entry, b_entry, c_entry, d_entry]))
        dirblocks.append((b"a", [e_entry, f_entry]))
        dirblocks.append((b"b", [g_entry, h_entry1, h_entry2, i_entry, j_entry]))
        state = dirstate.DirState.initialize("dirstate")
        state._validate()
        try:
            state._set_data([b"parent"], dirblocks)
        except:
            state.unlock()
            raise
        return state, dirblocks

    def test_iter_children_b(self):
        state, dirblocks = self.create_dirstate_with_two_trees()
        self.addCleanup(state.unlock)
        expected_result = []
        expected_result.append(dirblocks[3][1][2])  # h2
        expected_result.append(dirblocks[3][1][3])  # i
        expected_result.append(dirblocks[3][1][4])  # j
        self.assertEqual(expected_result, list(state._iter_child_entries(1, b"b")))

    def test_iter_child_root(self):
        state, dirblocks = self.create_dirstate_with_two_trees()
        self.addCleanup(state.unlock)
        expected_result = []
        expected_result.append(dirblocks[1][1][0])  # a
        expected_result.append(dirblocks[1][1][1])  # b
        expected_result.append(dirblocks[1][1][3])  # d
        expected_result.append(dirblocks[2][1][0])  # e
        expected_result.append(dirblocks[2][1][1])  # f
        expected_result.append(dirblocks[3][1][2])  # h2
        expected_result.append(dirblocks[3][1][3])  # i
        expected_result.append(dirblocks[3][1][4])  # j
        self.assertEqual(expected_result, list(state._iter_child_entries(1, b"")))


class InstrumentedDirState(dirstate.DirState):
    """An DirState with instrumented sha1 functionality."""

    def __init__(
        self,
        path,
        sha1_provider,
        worth_saving_limit=0,
        use_filesystem_for_exec=True,
        fdatasync=False,
    ):
        super().__init__(
            path,
            sha1_provider,
            worth_saving_limit=worth_saving_limit,
            use_filesystem_for_exec=use_filesystem_for_exec,
            fdatasync=fdatasync,
        )
        self._time_offset = 0
        self._log = []
        # member is dynamically set in DirState.__init__ to turn on trace
        self._sha1_provider = sha1_provider
        self._sha1_file = self._sha1_file_and_log

    def _sha_cutoff_time(self):
        timestamp = super()._sha_cutoff_time()
        self._cutoff_time = timestamp + self._time_offset

    def _sha1_file_and_log(self, abspath):
        self._log.append(("sha1", abspath))
        return self._sha1_provider.sha1(abspath)

    def _read_link(self, abspath, old_link):
        self._log.append(("read_link", abspath, old_link))
        return super()._read_link(abspath, old_link)

    def _lstat(self, abspath, entry):
        self._log.append(("lstat", abspath))
        return super()._lstat(abspath, entry)

    def _is_executable(self, mode, old_executable):
        self._log.append(("is_exec", mode, old_executable))
        return super()._is_executable(mode, old_executable)

    def adjust_time(self, secs):
        """Move the clock forward or back.

        :param secs: The amount to adjust the clock by. Positive values make it
        seem as if we are in the future, negative values make it seem like we
        are in the past.
        """
        self._time_offset += secs
        self._cutoff_time = None


class _FakeStat:
    """A class with the same attributes as a real stat result."""

    def __init__(self, size, mtime, ctime, dev, ino, mode):
        self.st_size = size
        self.st_mtime = mtime
        self.st_ctime = ctime
        self.st_dev = dev
        self.st_ino = ino
        self.st_mode = mode

    @staticmethod
    def from_stat(st):
        return _FakeStat(
            st.st_size, st.st_mtime, st.st_ctime, st.st_dev, st.st_ino, st.st_mode
        )


class TestDiscardMergeParents(TestCaseWithDirState, TestCaseInTempDir):
    def test_discard_no_parents(self):
        # This should be a no-op
        state = self.create_empty_dirstate()
        self.addCleanup(state.unlock)
        state._discard_merge_parents()
        state._validate()

    def test_discard_one_parent(self):
        # No-op
        packed_stat = b"AAAAREUHaIpFB2iKAAADAQAtkqUAAIGk"
        root_entry_direntry = (
            (b"", b"", b"a-root-value"),
            [
                (b"d", b"", 0, False, packed_stat),
                (b"d", b"", 0, False, packed_stat),
            ],
        )
        dirblocks = []
        dirblocks.append((b"", [root_entry_direntry]))
        dirblocks.append((b"", []))

        state = self.create_empty_dirstate()
        self.addCleanup(state.unlock)
        state._set_data([b"parent-id"], dirblocks[:])
        state._validate()

        state._discard_merge_parents()
        state._validate()
        self.assertEqual(dirblocks, state._dirblocks)

    def test_discard_simple(self):
        # No-op
        packed_stat = b"AAAAREUHaIpFB2iKAAADAQAtkqUAAIGk"
        root_entry_direntry = (
            (b"", b"", b"a-root-value"),
            [
                (b"d", b"", 0, False, packed_stat),
                (b"d", b"", 0, False, packed_stat),
                (b"d", b"", 0, False, packed_stat),
            ],
        )
        expected_root_entry_direntry = (
            (b"", b"", b"a-root-value"),
            [
                (b"d", b"", 0, False, packed_stat),
                (b"d", b"", 0, False, packed_stat),
            ],
        )
        dirblocks = []
        dirblocks.append((b"", [root_entry_direntry]))
        dirblocks.append((b"", []))

        state = self.create_empty_dirstate()
        self.addCleanup(state.unlock)
        state._set_data([b"parent-id", b"merged-id"], dirblocks[:])
        state._validate()

        # This should strip of the extra column
        state._discard_merge_parents()
        state._validate()
        expected_dirblocks = [(b"", [expected_root_entry_direntry]), (b"", [])]
        self.assertEqual(expected_dirblocks, state._dirblocks)

    def test_discard_absent(self):
        """If entries are only in a merge, discard should remove the entries."""
        null_stat = dirstate.DirState.NULLSTAT
        present_dir = (b"d", b"", 0, False, null_stat)
        present_file = (b"f", b"", 0, False, null_stat)
        absent = dirstate.DirState.NULL_PARENT_DETAILS
        root_key = (b"", b"", b"a-root-value")
        file_in_root_key = (b"", b"file-in-root", b"a-file-id")
        file_in_merged_key = (b"", b"file-in-merged", b"b-file-id")
        dirblocks = [
            (b"", [(root_key, [present_dir, present_dir, present_dir])]),
            (
                b"",
                [
                    (file_in_merged_key, [absent, absent, present_file]),
                    (file_in_root_key, [present_file, present_file, present_file]),
                ],
            ),
        ]

        state = self.create_empty_dirstate()
        self.addCleanup(state.unlock)
        state._set_data([b"parent-id", b"merged-id"], dirblocks[:])
        state._validate()

        exp_dirblocks = [
            (b"", [(root_key, [present_dir, present_dir])]),
            (
                b"",
                [
                    (file_in_root_key, [present_file, present_file]),
                ],
            ),
        ]
        state._discard_merge_parents()
        state._validate()
        self.assertEqual(exp_dirblocks, state._dirblocks)

    def test_discard_renamed(self):
        null_stat = dirstate.DirState.NULLSTAT
        present_dir = (b"d", b"", 0, False, null_stat)
        present_file = (b"f", b"", 0, False, null_stat)
        absent = dirstate.DirState.NULL_PARENT_DETAILS
        root_key = (b"", b"", b"a-root-value")
        file_in_root_key = (b"", b"file-in-root", b"a-file-id")
        # Renamed relative to parent
        file_rename_s_key = (b"", b"file-s", b"b-file-id")
        file_rename_t_key = (b"", b"file-t", b"b-file-id")
        # And one that is renamed between the parents, but absent in this
        key_in_1 = (b"", b"file-in-1", b"c-file-id")
        key_in_2 = (b"", b"file-in-2", b"c-file-id")

        dirblocks = [
            (b"", [(root_key, [present_dir, present_dir, present_dir])]),
            (
                b"",
                [
                    (
                        key_in_1,
                        [absent, present_file, (b"r", b"file-in-2", b"c-file-id")],
                    ),
                    (
                        key_in_2,
                        [absent, (b"r", b"file-in-1", b"c-file-id"), present_file],
                    ),
                    (file_in_root_key, [present_file, present_file, present_file]),
                    (
                        file_rename_s_key,
                        [(b"r", b"file-t", b"b-file-id"), absent, present_file],
                    ),
                    (
                        file_rename_t_key,
                        [present_file, absent, (b"r", b"file-s", b"b-file-id")],
                    ),
                ],
            ),
        ]
        exp_dirblocks = [
            (b"", [(root_key, [present_dir, present_dir])]),
            (
                b"",
                [
                    (key_in_1, [absent, present_file]),
                    (file_in_root_key, [present_file, present_file]),
                    (file_rename_t_key, [present_file, absent]),
                ],
            ),
        ]
        state = self.create_empty_dirstate()
        self.addCleanup(state.unlock)
        state._set_data([b"parent-id", b"merged-id"], dirblocks[:])
        state._validate()

        state._discard_merge_parents()
        state._validate()
        self.assertEqual(exp_dirblocks, state._dirblocks)

    def test_discard_all_subdir(self):
        null_stat = dirstate.DirState.NULLSTAT
        present_dir = (b"d", b"", 0, False, null_stat)
        present_file = (b"f", b"", 0, False, null_stat)
        absent = dirstate.DirState.NULL_PARENT_DETAILS
        root_key = (b"", b"", b"a-root-value")
        subdir_key = (b"", b"sub", b"dir-id")
        child1_key = (b"sub", b"child1", b"child1-id")
        child2_key = (b"sub", b"child2", b"child2-id")
        child3_key = (b"sub", b"child3", b"child3-id")

        dirblocks = [
            (b"", [(root_key, [present_dir, present_dir, present_dir])]),
            (b"", [(subdir_key, [present_dir, present_dir, present_dir])]),
            (
                b"sub",
                [
                    (child1_key, [absent, absent, present_file]),
                    (child2_key, [absent, absent, present_file]),
                    (child3_key, [absent, absent, present_file]),
                ],
            ),
        ]
        exp_dirblocks = [
            (b"", [(root_key, [present_dir, present_dir])]),
            (b"", [(subdir_key, [present_dir, present_dir])]),
            (b"sub", []),
        ]
        state = self.create_empty_dirstate()
        self.addCleanup(state.unlock)
        state._set_data([b"parent-id", b"merged-id"], dirblocks[:])
        state._validate()

        state._discard_merge_parents()
        state._validate()
        self.assertEqual(exp_dirblocks, state._dirblocks)


class Test_InvEntryToDetails(TestCase):
    def assertDetails(self, expected, inv_entry):
        details = dirstate._inv_entry_to_details(inv_entry)
        self.assertEqual(expected, details)
        # details should always allow join() and always be a plain str when
        # finished
        (minikind, fingerprint, _size, _executable, tree_data) = details
        self.assertIsInstance(minikind, bytes)
        self.assertIsInstance(fingerprint, bytes)
        self.assertIsInstance(tree_data, bytes)

    def test_unicode_symlink(self):
        target = "link-targ\N{EURO SIGN}t"
        inv_entry = inventory.InventoryLink(
            b"link-file-id",
            "nam\N{EURO SIGN}e",
            b"link-parent-id",
            b"link-revision-id",
            symlink_target=target,
        )
        self.assertDetails(
            (b"l", target.encode("UTF-8"), 0, False, b"link-revision-id"), inv_entry
        )


class TestSHA1Provider(TestCaseInTempDir):
    def test_sha1provider_is_an_interface(self):
        p = dirstate.SHA1Provider()
        self.assertRaises(NotImplementedError, p.sha1, "foo")
        self.assertRaises(NotImplementedError, p.stat_and_sha1, "foo")

    def test_defaultsha1provider_sha1(self):
        text = b"test\r\nwith\nall\rpossible line endings\r\n"
        self.build_tree_contents([("foo", text)])
        expected_sha = osutils.sha_string(text)
        p = dirstate.DefaultSHA1Provider()
        self.assertEqual(expected_sha, p.sha1("foo"))

    def test_defaultsha1provider_stat_and_sha1(self):
        text = b"test\r\nwith\nall\rpossible line endings\r\n"
        self.build_tree_contents([("foo", text)])
        expected_sha = osutils.sha_string(text)
        p = dirstate.DefaultSHA1Provider()
        statvalue, sha1 = p.stat_and_sha1("foo")
        self.assertEqual(len(text), statvalue.st_size)
        self.assertEqual(expected_sha, sha1)


class TestBisectDirblock(TestCase):
    """Test that bisect_dirblock() returns the expected values.

    bisect_dirblock is intended to work like bisect.bisect_left() except it
    knows it is working on dirblocks and that dirblocks are sorted by ('path',
    'to', 'foo') chunks rather than by raw 'path/to/foo'.
    """

    def assertBisect(self, dirblocks, split_dirblocks, path, *args, **kwargs):
        """Assert that bisect_split works like bisect_left on the split paths.

        :param dirblocks: A list of (path, [info]) pairs.
        :param split_dirblocks: A list of ((split, path), [info]) pairs.
        :param path: The path we are indexing.

        All other arguments will be passed along.
        """
        self.assertIsInstance(dirblocks, list)
        bisect_split_idx = dirstate.bisect_dirblock(dirblocks, path, *args, **kwargs)
        split_dirblock = (path.split(b"/"), [])
        bisect_left_idx = bisect.bisect_left(split_dirblocks, split_dirblock, *args)
        self.assertEqual(
            bisect_left_idx,
            bisect_split_idx,
            "bisect_split disagreed. {} != {} for key {!r}".format(
                bisect_left_idx, bisect_split_idx, path
            ),
        )

    def paths_to_dirblocks(self, paths):
        """Convert a list of paths into dirblock form.

        Also, ensure that the paths are in proper sorted order.
        """
        dirblocks = [(path, []) for path in paths]
        split_dirblocks = [(path.split(b"/"), []) for path in paths]
        self.assertEqual(sorted(split_dirblocks), split_dirblocks)
        return dirblocks, split_dirblocks

    def test_simple(self):
        """In the simple case it works just like bisect_left."""
        paths = [b"", b"a", b"b", b"c", b"d"]
        dirblocks, split_dirblocks = self.paths_to_dirblocks(paths)
        for path in paths:
            self.assertBisect(dirblocks, split_dirblocks, path)
        self.assertBisect(dirblocks, split_dirblocks, b"_")
        self.assertBisect(dirblocks, split_dirblocks, b"aa")
        self.assertBisect(dirblocks, split_dirblocks, b"bb")
        self.assertBisect(dirblocks, split_dirblocks, b"cc")
        self.assertBisect(dirblocks, split_dirblocks, b"dd")
        self.assertBisect(dirblocks, split_dirblocks, b"a/a")
        self.assertBisect(dirblocks, split_dirblocks, b"b/b")
        self.assertBisect(dirblocks, split_dirblocks, b"c/c")
        self.assertBisect(dirblocks, split_dirblocks, b"d/d")

    def test_involved(self):
        """This is where bisect_left diverges slightly."""
        paths = [
            b"",
            b"a",
            b"a/a",
            b"a/a/a",
            b"a/a/z",
            b"a/a-a",
            b"a/a-z",
            b"a/z",
            b"a/z/a",
            b"a/z/z",
            b"a/z-a",
            b"a/z-z",
            b"a-a",
            b"a-z",
            b"z",
            b"z/a/a",
            b"z/a/z",
            b"z/a-a",
            b"z/a-z",
            b"z/z",
            b"z/z/a",
            b"z/z/z",
            b"z/z-a",
            b"z/z-z",
            b"z-a",
            b"z-z",
        ]
        dirblocks, split_dirblocks = self.paths_to_dirblocks(paths)
        for path in paths:
            self.assertBisect(dirblocks, split_dirblocks, path)

    def test_involved_cached(self):
        """This is where bisect_left diverges slightly."""
        paths = [
            b"",
            b"a",
            b"a/a",
            b"a/a/a",
            b"a/a/z",
            b"a/a-a",
            b"a/a-z",
            b"a/z",
            b"a/z/a",
            b"a/z/z",
            b"a/z-a",
            b"a/z-z",
            b"a-a",
            b"a-z",
            b"z",
            b"z/a/a",
            b"z/a/z",
            b"z/a-a",
            b"z/a-z",
            b"z/z",
            b"z/z/a",
            b"z/z/z",
            b"z/z-a",
            b"z/z-z",
            b"z-a",
            b"z-z",
        ]
        cache = {}
        dirblocks, split_dirblocks = self.paths_to_dirblocks(paths)
        for path in paths:
            self.assertBisect(dirblocks, split_dirblocks, path, cache=cache)


def _unpack_stat(packed_stat):
    """Turn a packed_stat back into the stat fields.

    This is meant as a debugging tool, should not be used in real code.
    """
    (st_size, st_mtime, st_ctime, st_dev, st_ino, st_mode) = struct.unpack(
        ">6L", binascii.a2b_base64(packed_stat)
    )
    return {
        "st_size": st_size,
        "st_mtime": st_mtime,
        "st_ctime": st_ctime,
        "st_dev": st_dev,
        "st_ino": st_ino,
        "st_mode": st_mode,
    }


class TestPackStatRobust(TestCase):
    """Check packed representaton of stat values is robust on all inputs."""

    def pack(self, statlike_tuple):
        return dirstate.pack_stat(os.stat_result(statlike_tuple))

    @staticmethod
    def unpack_field(packed_string, stat_field):
        return _unpack_stat(packed_string)[stat_field]



# Copyright (C) 2025 Breezy Contributors
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Minimal transport for bzrformats.

Provides a Transport protocol and a simple in-memory implementation.
"""

import posixpath
from io import BytesIO
from typing import Protocol, runtime_checkable
from urllib.parse import unquote

from .errors import PathError


class NoSuchFile(PathError):
    """A file or directory does not exist."""

    _fmt = "No such file: %(path)r%(extra)s"


# Tuple for catching NoSuchFile from both bzrformats and breezy transports.
# Use this in except clauses when the transport may be either implementation.
try:
    from breezy.transport import NoSuchFile as _BreezyNoSuchFile

    TransportNoSuchFile = (NoSuchFile, _BreezyNoSuchFile)
except ImportError:
    TransportNoSuchFile = NoSuchFile


class FileExists(PathError):
    """A file or directory already exists."""

    _fmt = "File exists: %(path)r%(extra)s"


@runtime_checkable
class Transport(Protocol):
    """Minimal transport protocol for bzrformats."""

    base: str

    def get(self, relpath: str):
        """Get a file-like object for reading."""
        ...

    def get_bytes(self, relpath: str) -> bytes:
        """Get the raw bytes of a file."""
        ...

    def put_bytes(self, relpath: str, raw_bytes: bytes, mode=None):
        """Atomically put bytes at a location."""
        ...

    def put_file(self, relpath: str, f, mode=None) -> int:
        """Write a file from a file-like object, returning bytes written."""
        ...

    def put_file_non_atomic(self, relpath: str, f, mode=None, create_parent_dir=False):
        """Put a file-like object at a location."""
        ...

    def append_bytes(self, relpath: str, raw_bytes: bytes, mode=None) -> int:
        """Append bytes to a file, returning the byte offset of the start."""
        ...

    def readv(self, relpath: str, offsets):
        """Get parts of a file.

        :param offsets: List of (offset, size) tuples.
        :yields: (offset, data) tuples.
        """
        ...

    def open_write_stream(self, relpath: str, mode=None):
        """Open a writable stream at relpath."""
        ...

    def mkdir(self, relpath: str, mode=None):
        """Create a directory."""
        ...

    def delete(self, relpath: str):
        """Delete a file."""
        ...

    def move(self, rel_from: str, rel_to: str):
        """Move (rename) a file."""
        ...

    def stat(self, relpath: str):
        """Return a stat-like object for a file."""
        ...

    def has(self, relpath: str) -> bool:
        """Return True if the path exists."""
        ...

    def abspath(self, relpath: str) -> str:
        """Return the full URL for the given relative path."""
        ...

    def clone(self, relpath: str | None = None):
        """Return a new transport pointing at a sub-directory."""
        ...

    def iter_files_recursive(self):
        """Iterate over all files below this transport, yielding relpaths."""
        ...

    def ensure_base(self):
        """Ensure the base directory exists."""
        ...

    def recommended_page_size(self) -> int:
        """Return the recommended number of bytes to read at once."""
        ...


class _MemoryStat:
    """Minimal stat result for MemoryTransport."""

    def __init__(self, size, is_dir=False):
        self.st_size = size
        if is_dir:
            self.st_mode = 0o40755
        else:
            self.st_mode = 0o100644


class _MemoryWriteStream:
    """A write stream that writes directly to the backing store.

    Data is visible to readers immediately after each ``write()``.
    """

    def __init__(self, files, path):
        self._files = files
        self._path = path
        self._files.setdefault(path, b"")

    def write(self, data):
        self._files[self._path] = self._files.get(self._path, b"") + data

    def close(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()


def _sort_expand_and_combine(offsets, upper_limit, page_size):
    """Sort, expand, and combine readv offsets to reduce round trips.

    Each range is expanded to at least *page_size* bytes (centered on the
    original range), then overlapping ranges are merged.
    """
    if not offsets:
        return []
    sorted_offsets = sorted(offsets)
    expanded = []
    for offset, length in sorted_offsets:
        expansion = max(0, page_size - length)
        reduction = expansion // 2
        new_offset = max(0, offset - reduction)
        new_length = length + expansion
        if upper_limit:
            new_end = min(upper_limit, new_offset + new_length)
            new_length = max(0, new_end - new_offset)
        if new_length > 0:
            expanded.append((new_offset, new_length))
    if not expanded:
        return []
    merged = [expanded[0]]
    for offset, length in expanded[1:]:
        prev_offset, prev_length = merged[-1]
        prev_end = prev_offset + prev_length
        end = offset + length
        if offset > prev_end:
            merged.append((offset, length))
        elif end > prev_end:
            merged[-1] = (prev_offset, end - prev_offset)
    return merged


class MemoryTransport:
    """Simple in-memory transport for testing.

    All MemoryTransport instances sharing the same ``_files`` and ``_dirs``
    dicts see the same data, so :meth:`clone` produces a view onto the same
    store.
    """

    def __init__(self, url="memory:///", _files=None, _dirs=None):
        """Initialize MemoryTransport, optionally sharing an existing store."""
        if not url.endswith("/"):
            url += "/"
        self.base = url
        self._files = _files if _files is not None else {}
        self._dirs = _dirs if _dirs is not None else set()
        self._dirs.add("/")

    # -- internal helpers --

    def _abspath(self, relpath):
        """Resolve *relpath* to an absolute path within the store."""
        if relpath is None or relpath == ".":
            relpath = ""
        relpath = unquote(relpath)
        path = posixpath.join(self._path(), relpath)
        return posixpath.normpath(path)

    def _path(self):
        """Extract the path portion from the base URL."""
        path = self.base.split("://", 1)[-1]
        if path.endswith("/"):
            path = path[:-1]
        return path or "/"

    # -- Transport interface --

    def clone(self, relpath=None):
        """Return a new transport rooted at *relpath*."""
        if relpath is None:
            return MemoryTransport(self.base, self._files, self._dirs)
        return MemoryTransport(self.abspath(relpath), self._files, self._dirs)

    def abspath(self, relpath):
        """Return the full ``memory://`` URL for *relpath*."""
        return "memory://" + self._abspath(relpath)

    def has(self, relpath):
        """Return True if *relpath* exists as a file or directory."""
        path = self._abspath(relpath)
        return path in self._files or path in self._dirs

    def get(self, relpath):
        """Return a :class:`BytesIO` with the contents of *relpath*."""
        path = self._abspath(relpath)
        try:
            return BytesIO(self._files[path])
        except KeyError:
            raise NoSuchFile(relpath) from None

    def get_bytes(self, relpath):
        """Return the raw bytes of *relpath*."""
        path = self._abspath(relpath)
        try:
            return self._files[path]
        except KeyError:
            raise NoSuchFile(relpath) from None

    def put_bytes(self, relpath, raw_bytes, mode=None):
        """Store *raw_bytes* at *relpath*."""
        self._files[self._abspath(relpath)] = raw_bytes

    def put_file(self, relpath, f, mode=None):
        """Write *f* to *relpath*, returning the number of bytes written."""
        data = f.read()
        self._files[self._abspath(relpath)] = data
        return len(data)

    def put_file_non_atomic(self, relpath, f, mode=None, create_parent_dir=False):
        """Write *f* to *relpath*, creating parent dirs if requested."""
        if create_parent_dir:
            self._ensure_parent(relpath)
        self._files[self._abspath(relpath)] = f.read()

    def append_bytes(self, relpath, raw_bytes, mode=None):
        """Append *raw_bytes* to *relpath*, returning the start offset."""
        path = self._abspath(relpath)
        existing = self._files.get(path, b"")
        pos = len(existing)
        self._files[path] = existing + raw_bytes
        return pos

    def readv(self, relpath, offsets, adjust_for_latency=False, upper_limit=0):
        """Yield ``(offset, data)`` for each ``(offset, length)`` in *offsets*."""
        file_data = self.get_bytes(relpath)
        offsets = list(offsets)
        if adjust_for_latency and offsets:
            offsets = _sort_expand_and_combine(
                offsets, upper_limit or len(file_data), self.recommended_page_size()
            )
        for offset, length in offsets:
            yield offset, file_data[offset : offset + length]

    def open_write_stream(self, relpath, mode=None):
        """Return a writable stream; data is stored on close."""
        return _MemoryWriteStream(self._files, self._abspath(relpath))

    def mkdir(self, relpath, mode=None):
        """Create a directory at *relpath*.

        Does not raise if the directory already exists.
        """
        self._dirs.add(self._abspath(relpath))

    def delete(self, relpath):
        """Delete the file at *relpath*."""
        path = self._abspath(relpath)
        try:
            del self._files[path]
        except KeyError:
            raise NoSuchFile(relpath) from None

    def move(self, rel_from, rel_to):
        """Move (rename) a file from *rel_from* to *rel_to*."""
        path_from = self._abspath(rel_from)
        path_to = self._abspath(rel_to)
        try:
            self._files[path_to] = self._files.pop(path_from)
        except KeyError:
            raise NoSuchFile(rel_from) from None

    def stat(self, relpath):
        """Return a stat-like object for *relpath*."""
        path = self._abspath(relpath)
        if path in self._dirs:
            return _MemoryStat(0, is_dir=True)
        if path in self._files:
            return _MemoryStat(len(self._files[path]))
        raise NoSuchFile(relpath)

    def iter_files_recursive(self):
        """Yield relative paths of all files below this transport."""
        prefix = self._path().rstrip("/") + "/"
        for path in sorted(self._files):
            if path.startswith(prefix):
                yield path[len(prefix) :]

    def ensure_base(self):
        """Ensure the base directory exists."""
        self._dirs.add(self._path())

    def recommended_page_size(self):
        """Return a reasonable read-ahead size."""
        return 4096

    def _ensure_parent(self, relpath):
        """Ensure the parent directory of *relpath* exists."""
        parent = posixpath.dirname(self._abspath(relpath))
        self._dirs.add(parent)

    def __repr__(self):
        """Return string representation."""
        return f"MemoryTransport({self.base!r})"


class TracingTransport:
    """Transport wrapper that records operations in ``_activity``.

    Wraps another transport and delegates all calls. Selected operations
    are recorded as tuples in ``_activity`` for test assertions.  The
    tuple format matches breezy's ``TransportTraceDecorator``.
    """

    def __init__(self, inner):
        """Initialize with the transport to wrap."""
        self._inner = inner
        self._activity = []

    def __getattr__(self, name):
        """Delegate everything not explicitly overridden to the inner transport."""
        return getattr(self._inner, name)

    @property
    def base(self):
        """Return the base URL of the inner transport."""
        return self._inner.base

    # -- traced methods (match breezy's TransportTraceDecorator format) --

    def get(self, relpath):
        """Get file contents, recording the operation."""
        self._activity.append(("get", relpath))
        return self._inner.get(relpath)

    def get_bytes(self, relpath):
        """Get file bytes, recording the operation."""
        self._activity.append(("get", relpath))
        return self._inner.get_bytes(relpath)

    def put_bytes(self, relpath, raw_bytes, mode=None):
        """Put bytes, recording the operation."""
        self._activity.append(("put_bytes", relpath, len(raw_bytes), mode))
        return self._inner.put_bytes(relpath, raw_bytes, mode)

    def mkdir(self, relpath, mode=None):
        """Create a directory, recording the operation."""
        self._activity.append(("mkdir", relpath, mode))
        return self._inner.mkdir(relpath, mode)

    def readv(self, relpath, offsets, adjust_for_latency=False, upper_limit=None):
        """Read multiple ranges, recording the operation."""
        self._activity.append(
            ("readv", relpath, list(offsets), adjust_for_latency, upper_limit)
        )
        return self._inner.readv(
            relpath,
            offsets,
            adjust_for_latency=adjust_for_latency,
            upper_limit=upper_limit,
        )

    # -- non-traced pass-through for common methods --

    def put_file(self, relpath, f, mode=None):
        """Write a file to the inner transport."""
        return self._inner.put_file(relpath, f, mode)

    def clone(self, relpath=None):
        """Clone this tracing transport."""
        return TracingTransport(self._inner.clone(relpath))

    def recommended_page_size(self):
        """Return the recommended page size from the inner transport."""
        return self._inner.recommended_page_size()

    def __repr__(self):
        """Return string representation."""
        return f"TracingTransport({self._inner!r})"
